import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
)
from unstructured_ingest.connector.fsspec.sftp import SimpleSftpConfig

CMD_NAME = "sftp"


@dataclass
class SftpCliConfig(SimpleSftpConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--username"],
                required=True,
                type=str,
                help="Username for sftp connection",
            ),
            click.Option(
                ["--password"],
                required=True,
                type=str,
                help="Password for sftp connection",
            ),
            click.Option(
                ["--look-for-keys"],
                required=False,
                default=False,
                is_flag=True,
                type=bool,
                help="Whether to search for private key files in ~/.ssh/",
            ),
            click.Option(
                ["--allow-agent"],
                required=False,
                default=False,
                is_flag=True,
                type=bool,
                help="Whether to connect to the SSH agent.",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name=CMD_NAME,
        cli_config=SftpCliConfig,
        is_fsspec=True,
    )
    return cmd_cls



from __future__ import annotations

import os
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional
from urllib.parse import urlparse

from pydantic import Field, Secret

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.v2.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, FileDataSourceMetadata

if TYPE_CHECKING:
    from fsspec.implementations.sftp import SFTPFileSystem

CONNECTOR_TYPE = "sftp"


class SftpIndexerConfig(FsspecIndexerConfig):
    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        _, ext = os.path.splitext(self.remote_url)
        parsed_url = urlparse(self.remote_url)
        if ext:
            self.path_without_protocol = Path(parsed_url.path).parent.as_posix().lstrip("/")
        else:
            self.path_without_protocol = parsed_url.path.lstrip("/")


class SftpAccessConfig(FsspecAccessConfig):
    password: str = Field(description="Password for sftp connection")


class SftpConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = Field(default_factory=lambda: ["sftp"], init=False)
    access_config: Secret[SftpAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    username: str = Field(description="Username for sftp connection")
    host: Optional[str] = Field(default=None, description="Hostname for sftp connection")
    port: int = Field(default=22, description="Port for sftp connection")
    look_for_keys: bool = Field(
        default=False, description="Whether to search for private key files in ~/.ssh/"
    )
    allow_agent: bool = Field(default=False, description="Whether to connect to the SSH agent.")

    def get_access_config(self) -> dict[str, Any]:
        access_config = {
            "username": self.username,
            "host": self.host,
            "port": self.port,
            "look_for_keys": self.look_for_keys,
            "allow_agent": self.allow_agent,
            "password": self.access_config.get_secret_value().password,
        }
        return access_config

    @contextmanager
    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def get_client(self, protocol: str) -> Generator["SFTPFileSystem", None, None]:
        # The paramiko.SSHClient() client that's opened by the SFTPFileSystem
        # never gets closed so explicitly adding that as part of this context manager
        from fsspec import get_filesystem_class

        client: SFTPFileSystem = get_filesystem_class(protocol)(
            **self.get_access_config(),
        )
        yield client
        client.client.close()


@dataclass
class SftpIndexer(FsspecIndexer):
    connection_config: SftpConnectionConfig
    index_config: SftpIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def __post_init__(self):
        parsed_url = urlparse(self.index_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for file in super().run(**kwargs):
            new_identifier = (
                f"sftp://"
                f"{self.connection_config.host}:"
                f"{self.connection_config.port}/"
                f"{file.identifier}"
            )
            file.identifier = new_identifier
            yield file

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["name"]
        date_created = str(file_info.get("time").timestamp()) if "time" in file_info else None
        date_modified = str(file_info.get("mtime").timestamp()) if "mtime" in file_info else None

        file_size = file_info.get("size") if "size" in file_info else None

        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
        }
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class SftpDownloaderConfig(FsspecDownloaderConfig):
    remote_url: str = Field(description="Remote fsspec URL formatted as `protocol://dir/path`")


@dataclass
class SftpDownloader(FsspecDownloader):
    protocol: str = "sftp"
    connection_config: SftpConnectionConfig
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    download_config: Optional[SftpDownloaderConfig] = field(default_factory=SftpDownloaderConfig)

    def __post_init__(self):
        parsed_url = urlparse(self.download_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port


class SftpUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class SftpUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: SftpConnectionConfig
    upload_config: SftpUploaderConfig = field(default=None)


sftp_source_entry = SourceRegistryEntry(
    indexer=SftpIndexer,
    indexer_config=SftpIndexerConfig,
    downloader=SftpDownloader,
    downloader_config=SftpDownloaderConfig,
    connection_config=SftpConnectionConfig,
)

sftp_destination_entry = DestinationRegistryEntry(
    uploader=SftpUploader,
    uploader_config=SftpUploaderConfig,
    connection_config=SftpConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



import logging
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional, Tuple

import pandas as pd
from pydantic import Field, Secret

from unstructured_ingest.utils.data_prep import get_data_df
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.errors import ProviderError, UserAuthError, UserError
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    UploaderConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.sql.sql import (
    SQLUploader,
    SQLUploadStager,
    SQLUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData

if TYPE_CHECKING:
    from pyarrow import Table as ArrowTable
    from pyiceberg.catalog.rest import RestCatalog
    from pyiceberg.table import Table, Transaction

CONNECTOR_TYPE = "ibm_watsonx_s3"

DEFAULT_IBM_CLOUD_AUTH_URL = "https://iam.cloud.ibm.com/identity/token"
DEFAULT_ICEBERG_URI_PATH = "/mds/iceberg"
DEFAULT_ICEBERG_CATALOG_TYPE = "rest"


class IcebergCommitFailedException(Exception):
    """Failed to commit changes to the iceberg table."""


class IbmWatsonxAccessConfig(AccessConfig):
    iam_api_key: str = Field(description="IBM IAM API Key")
    access_key_id: str = Field(description="Cloud Object Storage HMAC Access Key ID")
    secret_access_key: str = Field(description="Cloud Object Storage HMAC Secret Access Key")


class IbmWatsonxConnectionConfig(ConnectionConfig):
    access_config: Secret[IbmWatsonxAccessConfig]
    iceberg_endpoint: str = Field(description="Iceberg REST endpoint")
    object_storage_endpoint: str = Field(description="Cloud Object Storage public endpoint")
    object_storage_region: str = Field(description="Cloud Object Storage region")
    catalog: str = Field(description="Catalog name")

    _bearer_token: Optional[dict[str, Any]] = None

    @property
    def iceberg_url(self) -> str:
        return f"https://{self.iceberg_endpoint.strip('/')}{DEFAULT_ICEBERG_URI_PATH}"

    @property
    def object_storage_url(self) -> str:
        return f"https://{self.object_storage_endpoint.strip('/')}"

    @property
    def bearer_token(self) -> str:
        # Add 60 seconds to deal with edge cases where the token expires before the request is made
        timestamp = int(time.time()) + 60
        if self._bearer_token is None or self._bearer_token.get("expiration", 0) <= timestamp:
            self._bearer_token = self.generate_bearer_token()
        return self._bearer_token["access_token"]

    @requires_dependencies(["httpx"], extras="ibm-watsonx-s3")
    def wrap_error(self, e: Exception) -> Exception:
        import httpx

        if not isinstance(e, httpx.HTTPStatusError):
            logger.error(f"Unhandled exception from IBM watsonx.data connector: {e}", exc_info=True)
            return e
        url = e.request.url
        response_code = e.response.status_code
        if response_code == 401:
            logger.error(
                f"Failed to authenticate IBM watsonx.data user {url}, status code {response_code}"
            )
            return UserAuthError(e)
        if response_code == 403:
            logger.error(
                f"Given IBM watsonx.data user is not authorized {url}, status code {response_code}"
            )
            return UserAuthError(e)
        if 400 <= response_code < 500:
            logger.error(
                f"Request to {url} failed"
                f"in IBM watsonx.data connector, status code {response_code}"
            )
            return UserError(e)
        if response_code > 500:
            logger.error(
                f"Request to {url} failed"
                f"in IBM watsonx.data connector, status code {response_code}"
            )
            return ProviderError(e)
        logger.error(f"Unhandled exception from IBM watsonx.data connector: {e}", exc_info=True)
        return e

    @requires_dependencies(["httpx"], extras="ibm-watsonx-s3")
    def generate_bearer_token(self) -> dict[str, Any]:
        import httpx

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        data = {
            "grant_type": "urn:ibm:params:oauth:grant-type:apikey",
            "apikey": self.access_config.get_secret_value().iam_api_key,
        }

        logger.info("Generating IBM IAM Bearer Token")
        try:
            response = httpx.post(DEFAULT_IBM_CLOUD_AUTH_URL, headers=headers, data=data)
            response.raise_for_status()
        except Exception as e:
            raise self.wrap_error(e)
        return response.json()

    def get_catalog_config(self) -> dict[str, Any]:
        return {
            "name": self.catalog,
            "type": DEFAULT_ICEBERG_CATALOG_TYPE,
            "uri": self.iceberg_url,
            "token": self.bearer_token,
            "warehouse": self.catalog,
            "s3.endpoint": self.object_storage_url,
            "s3.access-key-id": self.access_config.get_secret_value().access_key_id,
            "s3.secret-access-key": self.access_config.get_secret_value().secret_access_key,
            "s3.region": self.object_storage_region,
        }

    @requires_dependencies(["pyiceberg"], extras="ibm-watsonx-s3")
    @contextmanager
    def get_catalog(self) -> Generator["RestCatalog", None, None]:
        from pyiceberg.catalog import load_catalog

        try:
            catalog_config = self.get_catalog_config()
            catalog = load_catalog(**catalog_config)
        except Exception as e:
            logger.error(f"Failed to connect to catalog '{self.catalog}': {e}", exc_info=True)
            raise ProviderError(f"Failed to connect to catalog '{self.catalog}': {e}")

        yield catalog


@dataclass
class IbmWatsonxUploadStagerConfig(SQLUploadStagerConfig):
    pass


@dataclass
class IbmWatsonxUploadStager(SQLUploadStager):
    upload_stager_config: IbmWatsonxUploadStagerConfig = field(
        default_factory=IbmWatsonxUploadStagerConfig
    )


class IbmWatsonxUploaderConfig(UploaderConfig):
    namespace: str = Field(description="Namespace name")
    table: str = Field(description="Table name")
    max_retries: int = Field(
        default=5, description="Maximum number of retries to upload data", ge=2, le=10
    )
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="Searchable key to find entries for the same record on previous runs",
    )

    @property
    def table_identifier(self) -> Tuple[str, str]:
        return (self.namespace, self.table)


@dataclass
class IbmWatsonxUploader(SQLUploader):
    connection_config: IbmWatsonxConnectionConfig
    upload_config: IbmWatsonxUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        with self.connection_config.get_catalog() as catalog:
            if not catalog.namespace_exists(self.upload_config.namespace):
                raise UserError(f"Namespace '{self.upload_config.namespace}' does not exist")
            if not catalog.table_exists(self.upload_config.table_identifier):
                raise UserError(
                    f"Table '{self.upload_config.table}' does not exist in namespace '{self.upload_config.namespace}'"  # noqa: E501
                )

    @contextmanager
    def get_table(self) -> Generator["Table", None, None]:
        with self.connection_config.get_catalog() as catalog:
            table = catalog.load_table(self.upload_config.table_identifier)
            yield table

    def get_table_columns(self) -> list[str]:
        if self._columns is None:
            with self.get_table() as table:
                self._columns = table.schema().column_names
        return self._columns

    def can_delete(self) -> bool:
        return self.upload_config.record_id_key in self.get_table_columns()

    @requires_dependencies(["pyarrow"], extras="ibm-watsonx-s3")
    def _df_to_arrow_table(self, df: pd.DataFrame) -> "ArrowTable":
        import pyarrow as pa

        # Iceberg will automatically fill missing columns with nulls
        # Iceberg will throw an error if the DataFrame column has only null values
        # because it can't infer the type of the column and match it with the table schema
        return pa.Table.from_pandas(self._fit_to_schema(df, add_missing_columns=False))

    @requires_dependencies(["pyiceberg"], extras="ibm-watsonx-s3")
    def _delete(self, transaction: "Transaction", identifier: str) -> None:
        from pyiceberg.expressions import EqualTo

        if self.can_delete():
            transaction.delete(delete_filter=EqualTo(self.upload_config.record_id_key, identifier))
        else:
            logger.warning(
                f"Table doesn't contain expected "
                f"record id column "
                f"{self.upload_config.record_id_key}, skipping delete"
            )

    @requires_dependencies(["pyiceberg", "tenacity"], extras="ibm-watsonx-s3")
    def upload_data_table(
        self, table: "Table", data_table: "ArrowTable", file_data: FileData
    ) -> None:
        from pyiceberg.exceptions import CommitFailedException
        from tenacity import (
            before_log,
            retry,
            retry_if_exception_type,
            stop_after_attempt,
            wait_random,
        )

        @retry(
            stop=stop_after_attempt(self.upload_config.max_retries),
            wait=wait_random(),
            retry=retry_if_exception_type(IcebergCommitFailedException),
            before=before_log(logger, logging.DEBUG),
            reraise=True,
        )
        def _upload_data_table(table: "Table", data_table: "ArrowTable", file_data: FileData):
            try:
                with table.transaction() as transaction:
                    self._delete(transaction, file_data.identifier)
                    transaction.append(data_table)
            except CommitFailedException as e:
                table.refresh()
                logger.debug(e)
                raise IcebergCommitFailedException(e)
            except Exception as e:
                raise ProviderError(f"Failed to upload data to table: {e}")

        try:
            return _upload_data_table(table, data_table, file_data)
        except ProviderError:
            raise
        except Exception as e:
            raise ProviderError(f"Failed to upload data to table: {e}")

    def upload_dataframe(self, df: pd.DataFrame, file_data: FileData) -> None:
        data_table = self._df_to_arrow_table(df)

        with self.get_table() as table:
            self.upload_data_table(table, data_table, file_data)

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        df = pd.DataFrame(data)
        self.upload_dataframe(df=df, file_data=file_data)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path=path)
        self.upload_dataframe(df=df, file_data=file_data)


ibm_watsonx_s3_destination_entry = DestinationRegistryEntry(
    connection_config=IbmWatsonxConnectionConfig,
    uploader=IbmWatsonxUploader,
    uploader_config=IbmWatsonxUploaderConfig,
    upload_stager=IbmWatsonxUploadStager,
    upload_stager_config=IbmWatsonxUploadStagerConfig,
)



import json
import tempfile
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

import pandas as pd
import pytest
from _pytest.fixtures import TopRequest
from opensearchpy import Document, Keyword, OpenSearch, Text

from test.integration.connectors.utils.constants import DESTINATION_TAG, NOSQL_TAG, SOURCE_TAG
from test.integration.connectors.utils.docker import HealthCheck, container_context
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
)
from unstructured_ingest.v2.processes.connectors.elasticsearch.opensearch import (
    CONNECTOR_TYPE,
    OpenSearchAccessConfig,
    OpenSearchConnectionConfig,
    OpenSearchDownloader,
    OpenSearchDownloaderConfig,
    OpenSearchIndexer,
    OpenSearchIndexerConfig,
    OpenSearchUploader,
    OpenSearchUploaderConfig,
    OpenSearchUploadStager,
    OpenSearchUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

SOURCE_INDEX_NAME = "movies"
DESTINATION_INDEX_NAME = "elements"


class Movie(Document):
    title = Text(fields={"raw": Keyword()})
    year = Text()
    director = Text()
    cast = Text()
    genre = Text()
    wiki_page = Text()
    ethnicity = Text()
    plot = Text()

    class Index:
        name = SOURCE_INDEX_NAME

    def save(self, **kwargs):
        return super(Movie, self).save(**kwargs)


@contextmanager
def get_client() -> Generator[OpenSearch, None, None]:
    with OpenSearch(
        hosts=[{"host": "localhost", "port": 9200}],
        http_auth=("admin", "admin"),
        use_ssl=True,
        verify_certs=False,
        ssl_show_warn=False,
    ) as client:
        yield client


def get_index_count(client: OpenSearch, index_name: str) -> int:
    count_resp = client.cat.count(index=index_name, params={"format": "json"})
    return int(count_resp[0]["count"])


def wait_for_write(
    client: OpenSearch, index_name: str, expected_count: int, timeout: int = 30, interval: int = 1
) -> None:
    current_count = get_index_count(client, index_name)
    start = time.time()
    while time.time() - start < timeout:
        print(f"waiting for current count ({current_count}) to match expected {expected_count}")
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            return
    raise TimeoutError("Timed out while waiting for write to sync")


def validate_count(
    client: OpenSearch, index_name: str, expected_count: int, retries: int = 10, interval: int = 1
) -> None:
    current_count = get_index_count(client, index_name)
    if current_count == expected_count:
        return
    tries = 0
    while tries < retries:
        print(
            f"retrying validation to check if expected count "
            f"{expected_count} will match current count {current_count}"
        )
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            break
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from index: {current_count}"
    )


@pytest.fixture
def source_index(movies_dataframe: pd.DataFrame) -> str:
    with container_context(
        image="opensearchproject/opensearch:2.11.1",
        ports={9200: 9200, 9600: 9600},
        environment={"discovery.type": "single-node"},
        healthcheck=HealthCheck(
            test="curl --fail https://localhost:9200/_cat/health -ku 'admin:admin' >/dev/null || exit 1",  # noqa: E501
            interval=1,
        ),
    ):
        with get_client() as client:
            Movie.init(using=client)
            for i, row in movies_dataframe.iterrows():
                movie = Movie(
                    meta={"id": i},
                    title=row["Title"],
                    year=row["Release Year"],
                    director=row["Director"],
                    cast=row["Cast"],
                    genre=row["Genre"],
                    wiki_page=row["Wiki Page"],
                    ethnicity=row["Origin/Ethnicity"],
                    plot=row["Plot"],
                )
                movie.save(using=client)
            wait_for_write(
                client=client, index_name=SOURCE_INDEX_NAME, expected_count=len(movies_dataframe)
            )
        yield SOURCE_INDEX_NAME


@pytest.fixture
def destination_index(opensearch_elements_mapping: dict) -> str:
    with container_context(
        image="opensearchproject/opensearch:2.11.1",
        ports={9200: 9200, 9600: 9600},
        environment={"discovery.type": "single-node"},
        healthcheck=HealthCheck(
            test="curl --fail https://localhost:9200/_cat/health -ku 'admin:admin' >/dev/null || exit 1",  # noqa: E501
            interval=1,
        ),
    ):
        with get_client() as client:
            response = client.indices.create(
                index=DESTINATION_INDEX_NAME, body=opensearch_elements_mapping
            )
            if not response["acknowledged"]:
                raise RuntimeError(f"failed to create index: {response}")
        yield DESTINATION_INDEX_NAME


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
async def test_opensearch_source(source_index: str, movies_dataframe: pd.DataFrame):
    indexer_config = OpenSearchIndexerConfig(index_name=source_index)
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        connection_config = OpenSearchConnectionConfig(
            access_config=OpenSearchAccessConfig(password="admin"),
            username="admin",
            hosts=["http://localhost:9200"],
            use_ssl=True,
        )
        download_config = OpenSearchDownloaderConfig(download_dir=tempdir_path)
        indexer = OpenSearchIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = OpenSearchDownloader(
            connection_config=connection_config, download_config=download_config
        )
        expected_num_files = len(movies_dataframe)
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=SourceValidationConfigs(
                test_id=CONNECTOR_TYPE,
                expected_num_files=expected_num_files,
                expected_number_indexed_file_data=1,
                validate_downloaded_files=True,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
def test_opensearch_source_precheck_fail_no_cluster():
    indexer_config = OpenSearchIndexerConfig(index_name="index")

    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    indexer = OpenSearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
def test_opensearch_source_precheck_fail_no_index(source_index: str):
    indexer_config = OpenSearchIndexerConfig(index_name="index")

    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    indexer = OpenSearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
async def test_opensearch_destination(
    upload_file: Path,
    destination_index: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    stager = OpenSearchUploadStager(
        upload_stager_config=OpenSearchUploadStagerConfig(index_name=destination_index)
    )

    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name=destination_index),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
def test_opensearch_destination_precheck_fail():
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
def test_opensearch_destination_precheck_fail_no_index(destination_index: str):
    connection_config = OpenSearchConnectionConfig(
        access_config=OpenSearchAccessConfig(password="admin"),
        username="admin",
        hosts=["http://localhost:9200"],
        use_ssl=True,
    )
    uploader = OpenSearchUploader(
        connection_config=connection_config,
        upload_config=OpenSearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_opensearch_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = OpenSearchUploadStager(
        upload_stager_config=OpenSearchUploadStagerConfig(index_name="mock_index")
    )
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import json
import os
import time
from functools import lru_cache
from pathlib import Path
from typing import Generator
from uuid import uuid4

import pytest
import requests

from test.integration.connectors.utils.constants import DESTINATION_TAG, NOSQL_TAG
from test.integration.utils import requires_env
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.vectara import (
    CONNECTOR_TYPE as VECTARA_CONNECTOR_TYPE,
)
from unstructured_ingest.v2.processes.connectors.vectara import (
    VectaraAccessConfig,
    VectaraConnectionConfig,
    VectaraUploader,
    VectaraUploaderConfig,
    VectaraUploadStager,
    VectaraUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


def validate_upload(document: dict, expected_data: dict):
    logger.info(f"validating document: {document}")
    element_id = expected_data["element_id"]
    expected_text = expected_data["text"]
    filename = expected_data["metadata"]["filename"]
    filetype = expected_data["metadata"]["filetype"]
    page_number = expected_data["metadata"]["page_number"]

    assert document is not None
    speech_parts = document["parts"]
    assert speech_parts
    first_part = speech_parts[0]
    assert first_part["text"] == expected_text
    part_metadata = first_part["metadata"]
    assert part_metadata
    assert part_metadata["element_id"] == element_id
    assert part_metadata["filename"] == filename
    assert part_metadata["filetype"] == filetype
    assert part_metadata["page_number"] == page_number


@requires_env("VECTARA_OAUTH_CLIENT_ID", "VECTARA_OAUTH_SECRET", "VECTARA_CUSTOMER_ID")
@lru_cache()
def _get_jwt_token():
    """Connect to the server and get a JWT token."""
    customer_id = os.environ["VECTARA_CUSTOMER_ID"]
    token_endpoint = (
        f"https://vectara-prod-{customer_id}.auth.us-west-2.amazoncognito.com/oauth2/token"
    )
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    data = {
        "grant_type": "client_credentials",
        "client_id": os.environ["VECTARA_OAUTH_CLIENT_ID"],
        "client_secret": os.environ["VECTARA_OAUTH_SECRET"],
    }

    response = requests.post(token_endpoint, headers=headers, data=data)
    response.raise_for_status()
    response_json = response.json()

    return response_json.get("access_token")


def list_documents(corpus_key: str) -> list[str]:

    url = f"https://api.vectara.io/v2/corpora/{corpus_key}/documents"

    # the query below requires the corpus to have filter attributes for element_id

    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.get(url, headers=headers)
    response.raise_for_status()
    response_json = response.json()
    documents = response_json.get("documents", [])
    return documents


def fetch_document(corpus_key: str, documents_id: str) -> dict:
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}/documents/{documents_id}"
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.json()


def create_corpora(corpus_key: str, corpus_name: str) -> None:
    url = "https://api.vectara.io/v2/corpora"
    data = json.dumps({"key": corpus_key, "name": corpus_name, "description": "integration test"})
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.post(url, headers=headers, data=data)
    response.raise_for_status()


def replace_filter_attributes(corpus_key: str) -> None:
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}/replace_filter_attributes"
    data = json.dumps(
        {
            "filter_attributes": [
                {"name": "element_id", "level": "part", "indexed": True, "type": "text"}
            ]
        }
    )
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.post(url, headers=headers, data=data)
    response.raise_for_status()


def delete_corpora(corpus_key: str) -> None:
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}"

    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.delete(url, headers=headers)
    response.raise_for_status()


def get_metadata(corpus_key: str):
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}"
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.json()


def wait_for_ready(corpus_key: str, timeout=60, interval=2) -> None:
    start = time.time()
    while time.time() - start < timeout:
        try:
            get_metadata(corpus_key)
            return
        except requests.HTTPError:
            time.sleep(interval)
    raise TimeoutError("time out waiting for corpus to be ready")


def wait_for_delete(corpus_key: str, timeout=60, interval=2) -> None:
    start = time.time()
    while time.time() - start < timeout:
        try:
            get_metadata(corpus_key)
            time.sleep(interval)
        except requests.HTTPError:
            return
    raise TimeoutError("time out waiting for corpus to delete")


@pytest.fixture
def corpora_util() -> Generator[str, None, None]:
    random_id = str(uuid4()).split("-")[0]
    corpus_key = f"ingest-test-{random_id}"
    corpus_name = "ingest-test"
    logger.info(f"Creating corpus with key: {corpus_key}")
    try:
        create_corpora(corpus_key, corpus_name)
        replace_filter_attributes(corpus_key)
        wait_for_ready(corpus_key=corpus_key)
        yield corpus_key
    except Exception as e:
        logger.error(f"failed to create corpus {corpus_key}: {e}")
    finally:
        logger.info(f"deleting corpus: {corpus_key}")
        delete_corpora(corpus_key)
        wait_for_delete(corpus_key=corpus_key)


def wait_for_doc_meta(corpus_key: str, timeout=60, interval=1) -> list[str]:
    start = time.time()
    while time.time() - start < timeout:
        all_document_meta = list_documents(corpus_key)
        if not all_document_meta:
            time.sleep(interval)
            continue
        else:
            return all_document_meta
    raise TimeoutError("time out waiting for document to be ready")


@pytest.mark.asyncio
@pytest.mark.tags(VECTARA_CONNECTOR_TYPE, DESTINATION_TAG, "vectara", NOSQL_TAG)
@requires_env("VECTARA_OAUTH_CLIENT_ID", "VECTARA_OAUTH_SECRET", "VECTARA_CUSTOMER_ID")
async def test_vectara_destination(
    upload_file: Path, tmp_path: Path, corpora_util: str, retries=30, interval=1
):
    corpus_key = corpora_util
    connection_kwargs = {
        "customer_id": os.environ["VECTARA_CUSTOMER_ID"],
        "corpus_key": corpus_key,
    }

    oauth_client_id = os.environ["VECTARA_OAUTH_CLIENT_ID"]
    oauth_secret = os.environ["VECTARA_OAUTH_SECRET"]

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=VECTARA_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    stager_config = VectaraUploadStagerConfig()
    stager = VectaraUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=upload_file,
        output_dir=tmp_path,
        output_filename=upload_file.name,
        file_data=file_data,
    )

    uploader = VectaraUploader(
        connection_config=VectaraConnectionConfig(
            **connection_kwargs,
            access_config=VectaraAccessConfig(
                oauth_client_id=oauth_client_id, oauth_secret=oauth_secret
            ),
        ),
        upload_config=VectaraUploaderConfig(),
    )

    with new_upload_file.open() as new_upload_fp:
        elements_stager = json.load(new_upload_fp)

    if uploader.is_async():
        await uploader.run_data_async(data=elements_stager, file_data=file_data)

    with upload_file.open() as upload_fp:
        elements = json.load(upload_fp)
    first_element = elements[0]

    all_document_meta = wait_for_doc_meta(corpus_key)
    assert len(all_document_meta) == 1
    document_meta = all_document_meta[0]
    document = fetch_document(corpus_key=corpus_key, documents_id=document_meta["id"])
    validate_upload(document=document, expected_data=first_element)



from __future__ import annotations

import hashlib
from dataclasses import dataclass
from pathlib import Path
from time import time
from typing import Any, AsyncGenerator, Literal, Union

from pydantic import BaseModel, Field, Secret

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.html import HtmlMixin
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

from .client import ZendeskArticle, ZendeskClient, ZendeskTicket

CONNECTOR_TYPE = "zendesk"


class ZendeskAdditionalMetadata(BaseModel):
    item_type: Literal["ticket", "article"]
    content: Union[ZendeskTicket, ZendeskArticle]


class ZendeskFileData(FileData):
    additional_metadata: ZendeskAdditionalMetadata


class ZendeskAccessConfig(AccessConfig):
    api_token: str = Field(
        description="API token for zendesk generated under Apps and Integrations"
    )


class ZendeskConnectionConfig(ConnectionConfig):
    subdomain: str = Field(description="Subdomain for zendesk site, <sub-domain>.company.com")
    email: str = Field(description="Email for zendesk site registered at the subdomain")
    access_config: Secret[ZendeskAccessConfig]

    def get_client(self) -> ZendeskClient:

        access_config = self.access_config.get_secret_value()

        return ZendeskClient(
            email=self.email, subdomain=self.subdomain, token=access_config.api_token
        )


class ZendeskIndexerConfig(IndexerConfig):
    item_type: Literal["tickets", "articles", "all"] = Field(
        default="tickets",
        description="Type of item from zendesk to parse, can only be `tickets` or `articles`.",
    )


@dataclass
class ZendeskIndexer(Indexer):
    connection_config: ZendeskConnectionConfig
    index_config: ZendeskIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        """Validates connection to Zendesk API."""
        self.connection_config.get_client()

    def is_async(self) -> bool:
        return True

    def _generate_fullpath(self, identifier: str) -> Path:
        return Path(hashlib.sha256(identifier.encode("utf-8")).hexdigest()[:16] + ".txt")

    async def get_tickets(self) -> AsyncGenerator[ZendeskFileData, None]:
        async with self.connection_config.get_client() as client:
            async for ticket in client.get_tickets():
                yield ZendeskFileData(
                    identifier=str(ticket.id),
                    connector_type=self.connector_type,
                    source_identifiers=SourceIdentifiers(
                        filename=f"{ticket.id}.txt", fullpath=f"tickets/{ticket.id}.txt"
                    ),
                    additional_metadata=ZendeskAdditionalMetadata(
                        item_type="ticket", content=ticket
                    ),
                    metadata=FileDataSourceMetadata(
                        url=str(ticket.url) if ticket.url else None,
                        date_created=ticket.created_at.isoformat() if ticket.created_at else None,
                        date_modified=ticket.updated_at.isoformat() if ticket.updated_at else None,
                        date_processed=str(time()),
                    ),
                )

    async def get_articles(self) -> AsyncGenerator[ZendeskFileData, None]:
        async with self.connection_config.get_client() as client:
            async for article in client.get_articles():
                yield ZendeskFileData(
                    identifier=str(article.id),
                    connector_type=self.connector_type,
                    source_identifiers=SourceIdentifiers(
                        filename=f"{article.id}.html", fullpath=f"articles/{article.id}.html"
                    ),
                    additional_metadata=ZendeskAdditionalMetadata(
                        item_type="article", content=article
                    ),
                    metadata=FileDataSourceMetadata(
                        url=str(article.url) if article.url else None,
                        date_created=article.created_at.isoformat() if article.created_at else None,
                        date_modified=(
                            article.updated_at.isoformat() if article.updated_at else None
                        ),
                        date_processed=str(time()),
                    ),
                )

    async def run_async(self, **kwargs: Any) -> AsyncGenerator[ZendeskFileData, None]:
        """Determines item type and processes accordingly asynchronously."""
        item_type = self.index_config.item_type

        if item_type == "articles":
            async for article_file_data in self.get_articles():
                yield article_file_data

        elif item_type == "tickets":
            async for ticket_file_data in self.get_tickets():
                yield ticket_file_data

        elif item_type == "all":
            async for article_file_data in self.get_articles():
                yield article_file_data
            async for ticket_file_data in self.get_tickets():
                yield ticket_file_data

        else:
            raise ValueError(f"Item type {item_type} is not supported by the indexer")


class ZendeskDownloaderConfig(DownloaderConfig, HtmlMixin):
    pass


@dataclass
class ZendeskDownloader(Downloader):
    download_config: ZendeskDownloaderConfig
    connection_config: ZendeskConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def is_async(self) -> bool:
        return True

    def download_embedded_files(
        self, session, html: str, current_file_data: FileData
    ) -> list[DownloadResponse]:
        if not self.download_config.extract_files:
            return []
        url = current_file_data.metadata.url
        if url is None:
            logger.warning(
                f"""Missing URL for file: {current_file_data.source_identifiers.filename}.
                Skipping file extraction."""
            )
            return []
        filepath = current_file_data.source_identifiers.relative_path
        download_path = Path(self.download_dir) / filepath
        download_dir = download_path.with_suffix("")
        return self.download_config.extract_embedded_files(
            url=url,
            download_dir=download_dir,
            original_filedata=current_file_data,
            html=html,
            session=session,
        )

    @requires_dependencies(["aiofiles", "bs4"], extras="zendesk")
    async def download_article(self, article: ZendeskArticle, download_path: Path) -> None:
        import aiofiles
        import bs4

        article_html = article.as_html()
        soup = bs4.BeautifulSoup(article_html, "html.parser")
        async with aiofiles.open(download_path, "w", encoding="utf8") as f:
            await f.write(soup.prettify())

    @requires_dependencies(["aiofiles"], extras="zendesk")
    async def download_ticket(self, ticket: ZendeskTicket, download_path: Path) -> None:
        import aiofiles

        async with aiofiles.open(download_path, "w", encoding="utf8") as f:
            await f.write(ticket.as_text())
            async with self.connection_config.get_client() as client:
                comments = [comment async for comment in client.get_comments(ticket_id=ticket.id)]
                for comment in comments:
                    await f.write(comment.as_text())

    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:

        zendesk_filedata = ZendeskFileData.cast(file_data=file_data)

        item_type = zendesk_filedata.additional_metadata.item_type
        download_path = self.get_download_path(file_data=zendesk_filedata)
        download_path.parent.mkdir(parents=True, exist_ok=True)

        if item_type == "article":
            article = ZendeskArticle.model_validate(zendesk_filedata.additional_metadata.content)
            await self.download_article(article=article, download_path=download_path)
        elif item_type == "ticket":
            ticket = ZendeskTicket.model_validate(zendesk_filedata.additional_metadata.content)
            await self.download_ticket(ticket=ticket, download_path=download_path)
        else:
            raise RuntimeError(f"Item type {item_type} cannot be handled by the downloader")
        return super().generate_download_response(
            file_data=zendesk_filedata, download_path=download_path
        )


# create entry
zendesk_source_entry = SourceRegistryEntry(
    connection_config=ZendeskConnectionConfig,
    indexer_config=ZendeskIndexerConfig,
    indexer=ZendeskIndexer,
    downloader=ZendeskDownloader,
    downloader_config=ZendeskDownloaderConfig,
)



import json
from pathlib import Path

import pytest
from _pytest.fixtures import TopRequest
from psycopg2 import connect

from test.integration.connectors.utils.constants import (
    DESTINATION_TAG,
    SOURCE_TAG,
    SQL_TAG,
    env_setup_path,
)
from test.integration.connectors.utils.docker_compose import docker_compose_context
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.v2.processes.connectors.sql.postgres import (
    CONNECTOR_TYPE,
    PostgresAccessConfig,
    PostgresConnectionConfig,
    PostgresDownloader,
    PostgresDownloaderConfig,
    PostgresIndexer,
    PostgresIndexerConfig,
    PostgresUploader,
    PostgresUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

SEED_DATA_ROWS = 10


@pytest.fixture
def source_database_setup() -> str:
    db_name = "test_db"
    with docker_compose_context(docker_compose_path=env_setup_path / "sql" / "postgres" / "source"):
        connection = connect(
            user="unstructured",
            password="test",
            dbname=db_name,
            host="localhost",
            port=5433,
        )
        with connection.cursor() as cursor:
            for i in range(SEED_DATA_ROWS):
                sql_statment = f"INSERT INTO cars (brand, price) VALUES " f"('brand_{i}', {i})"
                cursor.execute(sql_statment)
            connection.commit()
        yield db_name


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, SQL_TAG)
async def test_postgres_source(temp_dir: Path, source_database_setup: str):
    connect_params = {
        "host": "localhost",
        "port": 5433,
        "database": "test_db",
        "user": "unstructured",
        "password": "test",
    }
    connection_config = PostgresConnectionConfig(
        host=connect_params["host"],
        port=connect_params["port"],
        database=connect_params["database"],
        username=connect_params["user"],
        access_config=PostgresAccessConfig(password=connect_params["password"]),
    )
    indexer = PostgresIndexer(
        connection_config=connection_config,
        index_config=PostgresIndexerConfig(table_name="cars", id_column="car_id", batch_size=6),
    )
    downloader = PostgresDownloader(
        connection_config=connection_config,
        download_config=PostgresDownloaderConfig(fields=["car_id", "brand"], download_dir=temp_dir),
    )
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="postgres",
            expected_num_files=SEED_DATA_ROWS,
            expected_number_indexed_file_data=2,
            validate_downloaded_files=True,
        ),
    )


def validate_destination(
    connect_params: dict,
    expected_num_elements: int,
    test_embedding: list[float],
    expected_text: str,
):
    # Run the following validations:
    # * Check that the number of records in the table match the expected value
    # * Given the embedding, make sure it matches the associated text it belongs to
    with connect(**connect_params) as connection:
        cursor = connection.cursor()
        query = "select count(*) from elements;"
        cursor.execute(query)
        count = cursor.fetchone()[0]
        assert (
            count == expected_num_elements
        ), f"dest check failed: got {count}, expected {expected_num_elements}"

        cursor.execute("SELECT embeddings FROM elements order by text limit 1")
        similarity_query = (
            f"SELECT text FROM elements ORDER BY embeddings <-> '{test_embedding}' LIMIT 1;"
        )
        cursor.execute(similarity_query)
        res = cursor.fetchone()
        assert res[0] == expected_text


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
async def test_postgres_destination(upload_file: Path, temp_dir: Path):
    # the postgres destination connector doesn't leverage the file data but is required as an input,
    # mocking it with arbitrary values to meet the base requirements:
    mock_file_data = FileData(
        identifier="mock file data",
        connector_type=CONNECTOR_TYPE,
        source_identifiers=SourceIdentifiers(filename=upload_file.name, fullpath=upload_file.name),
    )
    with docker_compose_context(
        docker_compose_path=env_setup_path / "sql" / "postgres" / "destination"
    ):
        stager = PostgresUploadStager()
        staged_path = stager.run(
            elements_filepath=upload_file,
            file_data=mock_file_data,
            output_dir=temp_dir,
            output_filename=upload_file.name,
        )

        # The stager should append the `.json` suffix to the output filename passed in.
        assert staged_path.suffix == upload_file.suffix

        connect_params = {
            "host": "localhost",
            "port": 5433,
            "database": "elements",
            "user": "unstructured",
            "password": "test",
        }

        uploader = PostgresUploader(
            connection_config=PostgresConnectionConfig(
                host=connect_params["host"],
                port=connect_params["port"],
                database=connect_params["database"],
                username=connect_params["user"],
                access_config=PostgresAccessConfig(password=connect_params["password"]),
            )
        )
        uploader.precheck()
        uploader.run(path=staged_path, file_data=mock_file_data)

        with staged_path.open("r") as f:
            staged_data = json.load(f)

        sample_element = staged_data[0]
        expected_num_elements = len(staged_data)
        validate_destination(
            connect_params=connect_params,
            expected_num_elements=expected_num_elements,
            expected_text=sample_element["text"],
            test_embedding=sample_element["embeddings"],
        )

        uploader.run(path=staged_path, file_data=mock_file_data)
        validate_destination(
            connect_params=connect_params,
            expected_num_elements=expected_num_elements,
            expected_text=sample_element["text"],
            test_embedding=sample_element["embeddings"],
        )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_postgres_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = PostgresUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import time
from contextlib import contextmanager
from typing import Optional, Union

import docker
from docker.models.containers import Container
from pydantic import BaseModel, Field, field_serializer


class HealthCheck(BaseModel):
    test: Union[str, list[str]]
    interval: int = Field(
        gt=0, default=30, description="The time to wait between checks in seconds."
    )
    timeout: int = Field(
        gt=0, default=30, description="The time to wait before considering the check to have hung."
    )
    retries: int = Field(
        gt=0,
        default=3,
        description="The number of consecutive failures needed "
        "to consider a container as unhealthy.",
    )
    start_period: int = Field(
        gt=0,
        default=0,
        description="Start period for the container to initialize before starting health-retries countdown in seconds.",  # noqa: E501
    )

    @field_serializer("interval")
    def serialize_interval(self, interval: int) -> int:
        return int(interval * 10e8)

    @field_serializer("timeout")
    def serialize_timeout(self, timeout: int) -> int:
        return int(timeout * 10e8)

    @field_serializer("start_period")
    def serialize_start_period(self, start_period: int) -> int:
        return int(start_period * 10e8)


def get_container(
    docker_client: docker.DockerClient,
    image: str,
    ports: dict,
    name: Optional[str] = None,
    environment: Optional[dict] = None,
    volumes: Optional[dict] = None,
    healthcheck: Optional[HealthCheck] = None,
) -> Container:
    run_kwargs = {
        "image": image,
        "detach": True,
        "ports": ports,
    }
    if environment:
        run_kwargs["environment"] = environment
    if volumes:
        run_kwargs["volumes"] = volumes
    if healthcheck:
        run_kwargs["healthcheck"] = healthcheck.model_dump()
    if name:
        run_kwargs["name"] = name
    container: Container = docker_client.containers.run(**run_kwargs)
    return container


def get_healthcheck(container: Container) -> Optional[HealthCheck]:
    healthcheck_config = container.attrs.get("Config", {}).get("Healthcheck", None)
    if not healthcheck_config:
        return None
    healthcheck_data = {
        "test": healthcheck_config["Test"],
    }
    if interval := healthcheck_config.get("Interval"):
        healthcheck_data["interval"] = interval / 10e8
    if start_period := healthcheck_config.get("StartPeriod"):
        healthcheck_data["start_period"] = start_period / 10e8
    if retries := healthcheck_config.get("Retries"):
        healthcheck_data["retries"] = retries
    return HealthCheck.model_validate(healthcheck_data)


def healthcheck_wait(
    container: Container, retries: int = 30, interval: int = 1, start_period: Optional[int] = None
) -> None:
    if start_period:
        time.sleep(start_period)
    health = container.health
    tries = 0
    while health != "healthy" and tries < retries:
        tries += 1
        logs = container.attrs.get("State", {}).get("Health", {}).get("Log")
        latest_log = logs[-1] if logs else None
        print(
            f"attempt {tries} - waiting for docker container "
            f"to be healthy: {health} latest log: {latest_log}"
        )
        time.sleep(interval)
        container.reload()
        health = container.health
    if health != "healthy":
        logs = container.attrs.get("State", {}).get("Health", {}).get("Log")
        latest_log = logs[-1] if logs else None
        raise TimeoutError(f"Docker container never came up healthy: {latest_log}")


@contextmanager
def container_context(
    image: str,
    ports: dict,
    environment: Optional[dict] = None,
    volumes: Optional[dict] = None,
    healthcheck: Optional[HealthCheck] = None,
    healthcheck_retries: int = 30,
    docker_client: Optional[docker.DockerClient] = None,
    name: Optional[str] = None,
):
    docker_client = docker_client or docker.from_env()
    print(f"pulling image {image}")
    docker_client.images.pull(image)
    container: Optional[Container] = None
    try:
        container = get_container(
            docker_client=docker_client,
            image=image,
            ports=ports,
            environment=environment,
            volumes=volumes,
            healthcheck=healthcheck,
            name=name,
        )
        if healthcheck_data := get_healthcheck(container):
            # Mirror whatever healthcheck config set on container
            healthcheck_wait(
                container=container,
                retries=healthcheck_retries,
                start_period=healthcheck_data.start_period,
                interval=healthcheck_data.interval,
            )
        yield container
    except AssertionError as e:
        if container:
            logs = container.logs()
            print(logs.decode("utf-8"))
            raise e
    finally:
        if container:
            container.kill()
            container.remove()



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional

from pydantic import Field, Secret

from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.utils.data_prep import split_dataframe
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.sql.sql import (
    SQLAccessConfig,
    SqlBatchFileData,
    SQLConnectionConfig,
    SQLDownloader,
    SQLDownloaderConfig,
    SQLIndexer,
    SQLIndexerConfig,
    SQLUploader,
    SQLUploaderConfig,
    SQLUploadStager,
    SQLUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import (
    FileData,
)
from unstructured_ingest.v2.utils import get_enhanced_element_id

if TYPE_CHECKING:
    from pandas import DataFrame
    from vastdb import connect as VastdbConnect
    from vastdb import transaction as VastdbTransaction
    from vastdb.table import Table as VastdbTable

CONNECTOR_TYPE = "vastdb"


class VastdbAccessConfig(SQLAccessConfig):
    endpoint: Optional[str] = Field(default=None, description="DB endpoint")
    access_key_id: Optional[str] = Field(default=None, description="access key id")
    access_key_secret: Optional[str] = Field(default=None, description="access key secret")


class VastdbConnectionConfig(SQLConnectionConfig):
    access_config: Secret[VastdbAccessConfig] = Field(
        default=VastdbAccessConfig(), validate_default=True
    )
    vastdb_bucket: str
    vastdb_schema: str
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @requires_dependencies(["vastdb"], extras="vastdb")
    @contextmanager
    def get_connection(self) -> "VastdbConnect":
        from vastdb import connect

        access_config = self.access_config.get_secret_value()
        connection = connect(
            endpoint=access_config.endpoint,
            access=access_config.access_key_id,
            secret=access_config.access_key_secret,
        )
        yield connection

    @contextmanager
    def get_cursor(self) -> "VastdbTransaction":
        with self.get_connection() as connection:
            with connection.transaction() as transaction:
                yield transaction

    @contextmanager
    def get_table(self, table_name: str) -> "VastdbTable":
        with self.get_cursor() as cursor:
            bucket = cursor.bucket(self.vastdb_bucket)
            schema = bucket.schema(self.vastdb_schema)
            table = schema.table(table_name)
            yield table


class VastdbIndexerConfig(SQLIndexerConfig):
    pass


@dataclass
class VastdbIndexer(SQLIndexer):
    connection_config: VastdbConnectionConfig
    index_config: VastdbIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def _get_doc_ids(self) -> list[str]:
        with self.connection_config.get_table(self.index_config.table_name) as table:
            reader = table.select(columns=[self.index_config.id_column])
            results = reader.read_all()  # Build a PyArrow Table from the RecordBatchReader
            ids = sorted([result[self.index_config.id_column] for result in results.to_pylist()])
            return ids

    def precheck(self) -> None:
        try:
            with self.connection_config.get_table(self.index_config.table_name) as table:
                table.select()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")


class VastdbDownloaderConfig(SQLDownloaderConfig):
    pass


@dataclass
class VastdbDownloader(SQLDownloader):
    connection_config: VastdbConnectionConfig
    download_config: VastdbDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["ibis"], extras="vastdb")
    def query_db(self, file_data: SqlBatchFileData) -> tuple[list[tuple], list[str]]:
        from ibis import _  # imports the Ibis deferred expression

        table_name = file_data.additional_metadata.table_name
        id_column = file_data.additional_metadata.id_column
        ids = tuple([item.identifier for item in file_data.batch_items])

        with self.connection_config.get_table(table_name) as table:
            predicate = _[id_column].isin(ids)

            if self.download_config.fields:
                # Vastdb requires the id column to be included in the fields
                fields = self.download_config.fields + [id_column]
                # dict.fromkeys to remove duplicates and keep order
                reader = table.select(columns=list(dict.fromkeys(fields)), predicate=predicate)
            else:
                reader = table.select(predicate=predicate)
            results = reader.read_all()
            df = results.to_pandas()
            return [tuple(r) for r in df.to_numpy()], results.column_names


class VastdbUploadStagerConfig(SQLUploadStagerConfig):
    rename_columns_map: Optional[dict] = Field(
        default=None,
        description="Map of column names to rename, ex: {'old_name': 'new_name'}",
    )


@dataclass
class VastdbUploadStager(SQLUploadStager):
    upload_stager_config: VastdbUploadStagerConfig = field(default_factory=VastdbUploadStagerConfig)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        metadata: dict[str, Any] = data.pop("metadata", {})
        data_source = metadata.pop("data_source", {})
        coordinates = metadata.pop("coordinates", {})

        data.update(metadata)
        data.update(data_source)
        data.update(coordinates)

        data["id"] = get_enhanced_element_id(element_dict=data, file_data=file_data)
        data[RECORD_ID_LABEL] = file_data.identifier
        return data

    def conform_dataframe(self, df: "DataFrame") -> "DataFrame":
        df = super().conform_dataframe(df=df)
        if self.upload_stager_config.rename_columns_map:
            df.rename(columns=self.upload_stager_config.rename_columns_map, inplace=True)
        return df


class VastdbUploaderConfig(SQLUploaderConfig):
    pass


@dataclass
class VastdbUploader(SQLUploader):
    upload_config: VastdbUploaderConfig = field(default_factory=VastdbUploaderConfig)
    connection_config: VastdbConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            with self.connection_config.get_table(self.upload_config.table_name) as table:
                table.select()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["pyarrow", "pandas"], extras="vastdb")
    def upload_dataframe(self, df: "DataFrame", file_data: FileData) -> None:
        import numpy as np
        import pyarrow as pa

        if self.can_delete():
            self.delete_by_record_id(file_data=file_data)
        else:
            logger.warning(
                f"table doesn't contain expected "
                f"record id column "
                f"{self.upload_config.record_id_key}, skipping delete"
            )
        df.replace({np.nan: None}, inplace=True)
        df = self._fit_to_schema(df=df)

        logger.info(
            f"writing a total of {len(df)} elements via"
            f" document batches to destination"
            f" table named {self.upload_config.table_name}"
            f" with batch size {self.upload_config.batch_size}"
        )

        for rows in split_dataframe(df=df, chunk_size=self.upload_config.batch_size):
            with self.connection_config.get_table(self.upload_config.table_name) as table:
                pa_table = pa.Table.from_pandas(rows)
                table.insert(pa_table)

    def get_table_columns(self) -> list[str]:
        if self._columns is None:
            with self.connection_config.get_table(self.upload_config.table_name) as table:
                self._columns = table.columns().names
        return self._columns

    @requires_dependencies(["ibis"], extras="vastdb")
    def delete_by_record_id(self, file_data: FileData) -> None:
        from ibis import _  # imports the Ibis deferred expression

        logger.debug(
            f"deleting any content with data "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from table {self.upload_config.table_name}"
        )
        predicate = _[self.upload_config.record_id_key].isin([file_data.identifier])
        with self.connection_config.get_table(self.upload_config.table_name) as table:
            # Get the internal row id
            rows_to_delete = table.select(
                columns=[], predicate=predicate, internal_row_id=True
            ).read_all()
            table.delete(rows_to_delete)


vastdb_source_entry = SourceRegistryEntry(
    connection_config=VastdbConnectionConfig,
    indexer_config=VastdbIndexerConfig,
    indexer=VastdbIndexer,
    downloader_config=VastdbDownloaderConfig,
    downloader=VastdbDownloader,
)

vastdb_destination_entry = DestinationRegistryEntry(
    connection_config=VastdbConnectionConfig,
    uploader=VastdbUploader,
    uploader_config=VastdbUploaderConfig,
    upload_stager=VastdbUploadStager,
    upload_stager_config=VastdbUploadStagerConfig,
)



import os
import uuid
from pathlib import Path
from typing import Generator

import duckdb
import pandas as pd
import pytest

from test.integration.connectors.utils.constants import DESTINATION_TAG, SQL_TAG
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.duckdb.motherduck import (
    CONNECTOR_TYPE,
    MotherDuckAccessConfig,
    MotherDuckConnectionConfig,
    MotherDuckUploader,
    MotherDuckUploaderConfig,
    MotherDuckUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def md_token() -> str:
    motherduck_token = os.getenv("MOTHERDUCK_TOKEN", None)
    assert motherduck_token
    return motherduck_token


@pytest.fixture
def provisioned_db(md_token: str, duckdb_schema: Path) -> Generator[str, None, None]:
    database_name = f"test_{str(uuid.uuid4()).replace('-', '_')}"
    try:
        with duckdb.connect(f"md:?motherduck_token={md_token}") as md_conn:
            with duckdb_schema.open("r") as f:
                query = f.read()
            md_conn.execute(f"CREATE DATABASE {database_name}")
            md_conn.execute(f"USE {database_name}")
            md_conn.execute(query)
            md_conn.close()
        yield database_name
    finally:
        with duckdb.connect(f"md:?motherduck_token={md_token}") as md_conn:
            md_conn.execute(f"DROP DATABASE {database_name}")
            md_conn.close()


def validate_motherduck_destination(database: str, expected_num_elements: int, md_token: str):
    conn = None
    try:
        conn = duckdb.connect(f"md:?motherduck_token={md_token}")
        conn.execute(f"USE {database}")
        _results = conn.sql("select count(*) from elements").fetchall()
        _count = _results[0][0]
        assert (
            _count == expected_num_elements
        ), f"dest check failed: got {_count}, expected {expected_num_elements}"
        conn.close()
    finally:
        if conn:
            conn.close()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
@requires_env("MOTHERDUCK_TOKEN")
def test_motherduck_destination(
    md_token: str, upload_file: Path, provisioned_db: str, temp_dir: Path
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    stager = MotherDuckUploadStager()
    staged_path = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=temp_dir,
        output_filename=upload_file.name,
    )

    access_config = MotherDuckAccessConfig(md_token=md_token)
    connection_config = MotherDuckConnectionConfig(
        database=provisioned_db, access_config=access_config
    )
    upload_config = MotherDuckUploaderConfig()
    uploader = MotherDuckUploader(connection_config=connection_config, upload_config=upload_config)

    uploader.run(path=staged_path, file_data=file_data)

    staged_df = pd.read_json(staged_path, orient="records", lines=True)
    validate_motherduck_destination(
        database=provisioned_db, expected_num_elements=len(staged_df), md_token=md_token
    )



import time
from unittest.mock import MagicMock

import pandas as pd
import pytest
from pydantic import Secret
from pyiceberg.exceptions import CommitFailedException
from pytest_mock import MockerFixture

from unstructured_ingest.v2.errors import ProviderError, UserError
from unstructured_ingest.v2.processes.connectors.ibm_watsonx import IBM_WATSONX_S3_CONNECTOR_TYPE
from unstructured_ingest.v2.processes.connectors.ibm_watsonx.ibm_watsonx_s3 import (
    IbmWatsonxAccessConfig,
    IbmWatsonxConnectionConfig,
    IbmWatsonxUploader,
    IbmWatsonxUploaderConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def file_data():
    return FileData(
        identifier="test_identifier",
        connector_type=IBM_WATSONX_S3_CONNECTOR_TYPE,
        source_identifiers=SourceIdentifiers(
            filename="test_file.pdf", fullpath="/tmp/test_file.pdf"
        ),
    )


@pytest.fixture
def access_config():
    return IbmWatsonxAccessConfig(
        iam_api_key="test_iam_api_key",
        access_key_id="test_access_key_id",
        secret_access_key="test_secret_access_key",
    )


@pytest.fixture
def connection_config(access_config: IbmWatsonxAccessConfig):
    return IbmWatsonxConnectionConfig(
        access_config=Secret(access_config),
        iceberg_endpoint="test_iceberg_endpoint/",
        object_storage_endpoint="test_object_storage_endpoint/",
        object_storage_region="test_region",
        catalog="test_catalog",
    )


@pytest.fixture
def uploader_config():
    return IbmWatsonxUploaderConfig(
        namespace="test_namespace",
        table="test_table",
        record_id_key="test_record_id_key",
    )


@pytest.fixture
def uploader(
    connection_config: IbmWatsonxConnectionConfig, uploader_config: IbmWatsonxUploaderConfig
):
    return IbmWatsonxUploader(
        connection_config=connection_config,
        upload_config=uploader_config,
    )


@pytest.fixture
def mock_catalog(mocker: MockerFixture):
    mock_catalog = mocker.MagicMock()
    mock_catalog.namespace_exists.return_value = True
    mock_catalog.table_exists.return_value = True
    return mock_catalog


@pytest.fixture
def mock_get_catalog(mocker: MockerFixture, mock_catalog: MagicMock):
    mock_get_catalog = mocker.patch.context_manager(
        IbmWatsonxConnectionConfig, "get_catalog", autospec=True
    )
    mock_get_catalog.return_value.__enter__.return_value = mock_catalog
    return mock_get_catalog


@pytest.fixture
def mock_table(mocker: MockerFixture):
    mock_table = mocker.MagicMock()
    return mock_table


@pytest.fixture
def mock_get_table(mocker: MockerFixture, mock_table: MagicMock):
    mock_get_table = mocker.patch.context_manager(IbmWatsonxUploader, "get_table", autospec=True)
    mock_get_table.return_value.__enter__.return_value = mock_table
    return mock_get_table


@pytest.fixture
def mock_transaction(mocker: MockerFixture, mock_table: MagicMock):
    mock_transaction = mocker.MagicMock()
    mock_table.transaction.return_value.__enter__.return_value = mock_transaction
    return mock_transaction


@pytest.fixture
def mock_data_table(mocker: MockerFixture):
    mock_data_table = mocker.MagicMock()
    mock_data_table.schema = "schema"
    return mock_data_table


@pytest.fixture
def mock_delete(mocker: MockerFixture):
    return mocker.patch.object(IbmWatsonxUploader, "_delete")


@pytest.fixture
def test_df():
    return pd.DataFrame(
        {
            "test_column_0": [True, False, True],
            "test_column_1": [1, 2, 3],
            "test_column_2": ["a", "b", "c"],
        }
    )


@pytest.fixture
def timestamp_now():
    return int(time.time())


def test_ibm_watsonx_connection_config_iceberg_url(
    mocker: MockerFixture,
    connection_config: IbmWatsonxConnectionConfig,
):
    mocker.patch(
        "unstructured_ingest.v2.processes.connectors.ibm_watsonx.ibm_watsonx_s3.DEFAULT_ICEBERG_URI_PATH",  # noqa: E501
        new="/mds/iceberg",
    )
    expected_url = "https://test_iceberg_endpoint/mds/iceberg"
    assert connection_config.iceberg_url == expected_url


def test_ibm_watsonx_connection_config_object_storage_url(
    connection_config: IbmWatsonxConnectionConfig,
):
    expected_url = "https://test_object_storage_endpoint"
    assert connection_config.object_storage_url == expected_url


def test_ibm_watsonx_connection_config_bearer_token_new_token(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig, timestamp_now: int
):
    mock_generate_bearer_token = mocker.patch.object(
        IbmWatsonxConnectionConfig,
        "generate_bearer_token",
        return_value={"access_token": "new_token", "expiration": timestamp_now + 3600},
    )
    token = connection_config.bearer_token
    assert token == "new_token"
    mock_generate_bearer_token.assert_called_once()


def test_ibm_watsonx_connection_config_bearer_token_existing_token(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig, timestamp_now: int
):
    connection_config._bearer_token = {
        "access_token": "existing_token",
        "expiration": timestamp_now + 3600,
    }
    mock_generate_bearer_token = mocker.patch.object(
        IbmWatsonxConnectionConfig, "generate_bearer_token"
    )
    token = connection_config.bearer_token
    assert token == "existing_token"
    mock_generate_bearer_token.assert_not_called()


def test_ibm_watsonx_connection_config_bearer_token_expired_token(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig, timestamp_now: int
):
    connection_config._bearer_token = {
        "access_token": "expired_token",
        "expiration": timestamp_now - 3600,
    }
    mock_generate_bearer_token = mocker.patch.object(
        IbmWatsonxConnectionConfig,
        "generate_bearer_token",
        return_value={"access_token": "new_token", "expiration": timestamp_now + 3600},
    )
    token = connection_config.bearer_token
    assert token == "new_token"
    mock_generate_bearer_token.assert_called_once()


def test_ibm_watsonx_connection_config_bearer_token_soon_to_expire_token(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig, timestamp_now: int
):
    connection_config._bearer_token = {
        "access_token": "soon_to_expire_token",
        "expiration": timestamp_now + 60,
    }
    mock_generate_bearer_token = mocker.patch.object(
        IbmWatsonxConnectionConfig,
        "generate_bearer_token",
        return_value={"access_token": "new_token", "expiration": timestamp_now + 3600},
    )
    token = connection_config.bearer_token
    assert token == "new_token"
    mock_generate_bearer_token.assert_called_once()


def test_ibm_watsonx_connection_config_get_catalog_success(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig
):
    mocker.patch(
        "unstructured_ingest.v2.processes.connectors.ibm_watsonx.ibm_watsonx_s3.DEFAULT_ICEBERG_URI_PATH",  # noqa: E501
        new="/mds/iceberg",
    )
    mocker.patch.object(
        IbmWatsonxConnectionConfig,
        "bearer_token",
        new="test_bearer_token",
    )
    mock_load_catalog = mocker.patch("pyiceberg.catalog.load_catalog")

    with connection_config.get_catalog() as catalog:
        assert catalog is not None
    mock_load_catalog.assert_called_once_with(
        **{
            "name": "test_catalog",
            "type": "rest",
            "uri": "https://test_iceberg_endpoint/mds/iceberg",
            "token": "test_bearer_token",
            "warehouse": "test_catalog",
            "s3.endpoint": "https://test_object_storage_endpoint",
            "s3.access-key-id": "test_access_key_id",
            "s3.secret-access-key": "test_secret_access_key",
            "s3.region": "test_region",
        }
    )


def test_ibm_watsonx_connection_config_get_catalog_failure(
    mocker: MockerFixture, connection_config: IbmWatsonxConnectionConfig
):
    mocker.patch(
        "pyiceberg.catalog.load_catalog",
        side_effect=Exception("Connection error"),
    )
    mocker.patch.object(
        IbmWatsonxConnectionConfig,
        "bearer_token",
        new="test_bearer_token",
    )
    with pytest.raises(ProviderError):
        with connection_config.get_catalog():
            pass


def test_ibm_watsonx_uploader_precheck_namespace_exists_table_exists(
    mock_get_catalog: MagicMock,
    mock_catalog: MagicMock,
    uploader: IbmWatsonxUploader,
):
    uploader.precheck()

    mock_catalog.namespace_exists.assert_called_once_with("test_namespace")
    mock_catalog.table_exists.assert_called_once_with(("test_namespace", "test_table"))


def test_ibm_watsonx_uploader_precheck_namespace_does_not_exist(
    mock_get_catalog: MagicMock,
    mock_catalog: MagicMock,
    uploader: IbmWatsonxUploader,
):
    mock_catalog.namespace_exists.return_value = False

    with pytest.raises(UserError, match="Namespace 'test_namespace' does not exist"):
        uploader.precheck()

    mock_catalog.namespace_exists.assert_called_once_with("test_namespace")
    mock_catalog.table_exists.assert_not_called()


def test_ibm_watsonx_uploader_precheck_table_does_not_exist(
    mock_get_catalog: MagicMock,
    mock_catalog: MagicMock,
    uploader: IbmWatsonxUploader,
):
    mock_catalog.table_exists.return_value = False

    with pytest.raises(
        UserError,
        match="Table 'test_table' does not exist in namespace 'test_namespace'",
    ):
        uploader.precheck()

    mock_catalog.namespace_exists.assert_called_once_with("test_namespace")
    mock_catalog.table_exists.assert_called_once_with(("test_namespace", "test_table"))


def test_ibm_watsonx_uploader_upload_data_table_success(
    uploader: IbmWatsonxUploader,
    mock_table: MagicMock,
    mock_transaction: MagicMock,
    mock_data_table: MagicMock,
    mock_delete: MagicMock,
    file_data: FileData,
):
    uploader.upload_data_table(mock_table, mock_data_table, file_data)

    mock_delete.assert_called_once_with(mock_transaction, "test_identifier")
    mock_transaction.append.assert_called_once_with(mock_data_table)


def test_ibm_watsonx_uploader_upload_data_table_commit_exception(
    uploader: IbmWatsonxUploader,
    mock_table: MagicMock,
    mock_transaction: MagicMock,
    mock_data_table: MagicMock,
    mock_delete: MagicMock,
    file_data: FileData,
):
    mock_transaction.append.side_effect = CommitFailedException()

    with pytest.raises(ProviderError):
        uploader.upload_data_table(mock_table, mock_data_table, file_data)
    assert mock_table.refresh.call_count == 5


def test_ibm_watsonx_uploader_upload_data_table_exception(
    uploader: IbmWatsonxUploader,
    mock_table: MagicMock,
    mock_transaction: MagicMock,
    mock_data_table: MagicMock,
    mock_delete: MagicMock,
    file_data: FileData,
):
    mock_transaction.append.side_effect = Exception()

    with pytest.raises(ProviderError):
        uploader.upload_data_table(mock_table, mock_data_table, file_data)
    assert mock_table.refresh.call_count == 0


def test_ibm_watsonx_uploader_df_to_arrow_table(
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
    test_df: pd.DataFrame,
):
    mock_fit_to_schema = mocker.patch.object(
        IbmWatsonxUploader, "_fit_to_schema", return_value=test_df
    )

    result = uploader._df_to_arrow_table(test_df)

    mock_fit_to_schema.assert_called_once_with(test_df, add_missing_columns=False)
    assert len(result.column_names) == 3
    assert "test_column_0" in result.column_names
    assert "test_column_1" in result.column_names
    assert "test_column_2" in result.column_names


def test_ibm_watsonx_uploader_can_delete_column_exists(
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
):
    mocker.patch.object(
        IbmWatsonxUploader, "get_table_columns", return_value=["test_record_id_key"]
    )

    assert uploader.can_delete() is True


def test_ibm_watsonx_uploader_can_delete_column_does_not_exist(
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
):
    mocker.patch.object(IbmWatsonxUploader, "get_table_columns", return_value=["other_column"])

    assert uploader.can_delete() is False


def test_ibm_watsonx_uploader_get_table_columns_cache(
    uploader: IbmWatsonxUploader,
):
    uploader._columns = ["cached_column"]

    result = uploader.get_table_columns()

    assert result == ["cached_column"]


def test_ibm_watsonx_uploader_get_table_columns_no_cache(
    uploader: IbmWatsonxUploader,
    mock_get_table: MagicMock,
    mock_table: MagicMock,
):
    uploader._columns = None
    mock_table.schema.return_value.column_names = ["column_1", "column_2"]

    result = uploader.get_table_columns()

    mock_get_table.assert_called_once()
    assert result == ["column_1", "column_2"]
    assert uploader._columns == ["column_1", "column_2"]


def test_ibm_watsonx_uploader_upload_dataframe_success(
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
    test_df: pd.DataFrame,
    mock_get_table: MagicMock,
    mock_table: MagicMock,
    mock_data_table: MagicMock,
    file_data: FileData,
):
    mocker.patch.object(IbmWatsonxUploader, "_df_to_arrow_table", return_value=mock_data_table)
    mock_upload_data_table = mocker.patch.object(IbmWatsonxUploader, "upload_data_table")

    uploader.upload_dataframe(test_df, file_data)

    mock_get_table.assert_called_once()
    mock_upload_data_table.assert_called_once_with(mock_table, mock_data_table, file_data)


def test_ibm_watsonx_uploader_delete_can_delete(
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
    mock_transaction: MagicMock,
):
    mocker.patch.object(IbmWatsonxUploader, "can_delete", return_value=True)
    mock_equal_to = mocker.patch("pyiceberg.expressions.EqualTo")

    uploader._delete(mock_transaction, "test_identifier")

    mock_equal_to.assert_called_once_with("test_record_id_key", "test_identifier")
    mock_transaction.delete.assert_called_once_with(delete_filter=mock_equal_to.return_value)


def test_ibm_watsonx_uploader_delete_cannot_delete(
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
    uploader: IbmWatsonxUploader,
    mock_transaction: MagicMock,
):
    mocker.patch.object(IbmWatsonxUploader, "can_delete", return_value=False)

    uploader._delete(mock_transaction, "test_identifier")
    mock_transaction.delete.assert_not_called()
    assert (
        "Table doesn't contain expected record id column test_record_id_key, skipping delete"
        in caplog.text
    )



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
)
from unstructured_ingest.connector.fsspec.azure import (
    AzureWriteConfig,
    SimpleAzureBlobStorageConfig,
)

CMD_NAME = "azure"


@dataclass
class AzureCliConfig(SimpleAzureBlobStorageConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--account-key"],
                default=None,
                help="The storage account key. This is used for shared key "
                "authentication. If any of account key, sas token or "
                "client_id are not specified, anonymous access will be used.",
            ),
            click.Option(
                ["--account-name"],
                default=None,
                help="The storage account name. This is used to authenticate "
                "requests signed with an account key and to construct "
                "the storage endpoint. It is required unless a connection "
                "string is given, or if a custom domain is used with "
                "anonymous authentication.",
            ),
            click.Option(
                ["--connection-string"],
                default=None,
                help="If specified, this will override all other parameters. See "
                "http://azure.microsoft.com/en-us/documentation/articles/storage-configure-connection-string/ "  # noqa: E501
                "for the connection string format.",
            ),
            click.Option(
                ["--sas_token"],
                default=None,
                help="A shared access signature token to use to authenticate "
                "requests instead of the account key. If account key and "
                "sas token are both specified, account key will be used "
                "to sign. If any of account key, sas token or client_id "
                "are not specified, anonymous access will be used.",
            ),
        ]
        return options


@dataclass
class AzureCliWriteConfig(AzureWriteConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--overwrite"],
                is_flag=True,
                default=False,
                show_default=True,
                help="If set, will overwrite content if content already exists",
            )
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name=CMD_NAME,
        cli_config=AzureCliConfig,
        is_fsspec=True,
    )
    return cmd_cls


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name=CMD_NAME,
        cli_config=AzureCliConfig,
        write_config=AzureCliWriteConfig,
        is_fsspec=True,
        additional_cli_options=[AzureCliWriteConfig],
    )
    return cmd_cls



# ruff: noqa: I001
import json
import tempfile
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Generator
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
import pandas as pd
import pytest
from _pytest.fixtures import TopRequest
from elasticsearch import Elasticsearch as ElasticsearchClient
from elasticsearch.helpers import bulk

from test.integration.connectors.utils.constants import DESTINATION_TAG, SOURCE_TAG, NOSQL_TAG
from test.integration.connectors.utils.docker import HealthCheck, container_context
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.elasticsearch.elasticsearch import (
    CONNECTOR_TYPE,
    ElasticsearchAccessConfig,
    ElasticsearchConnectionConfig,
    ElasticsearchDownloader,
    ElasticsearchDownloaderConfig,
    ElasticsearchIndexer,
    ElasticsearchIndexerConfig,
    ElasticsearchUploader,
    ElasticsearchUploaderConfig,
    ElasticsearchUploadStager,
    ElasticsearchUploadStagerConfig,
)

SOURCE_INDEX_NAME = "movies"
DESTINATION_INDEX_NAME = "elements"
ES_USERNAME = "elastic"
ES_PASSWORD = "elastic_password"


@contextmanager
def get_client() -> Generator[ElasticsearchClient, None, None]:
    with ElasticsearchClient(
        hosts="http://localhost:9200", basic_auth=(ES_USERNAME, ES_PASSWORD), request_timeout=30
    ) as client:
        yield client


def form_elasticsearch_doc_dict(i, csv_row):
    return {
        "_index": SOURCE_INDEX_NAME,
        "_id": i,
        "_source": {
            "title": csv_row["Title"],
            "ethnicity": csv_row["Origin/Ethnicity"],
            "director": csv_row["Director"],
            "cast": csv_row["Cast"],
            "genre": csv_row["Genre"],
            "plot": csv_row["Plot"],
            "year": csv_row["Release Year"],
            "wiki_page": csv_row["Wiki Page"],
        },
    }


def dataframe_to_upload_data(df: pd.DataFrame) -> list[dict]:
    upload_data = []
    for index, row in df.iterrows():
        upload_data.append(form_elasticsearch_doc_dict(index, row))
    return upload_data


def get_index_count(client: ElasticsearchClient, index_name: str) -> int:
    count_resp = client.cat.count(index=index_name, format="json")
    return int(count_resp[0]["count"])


def validate_count(
    client: ElasticsearchClient,
    index_name: str,
    expected_count: int,
    retries: int = 10,
    interval: int = 1,
) -> None:
    current_count = get_index_count(client, index_name)
    if current_count == expected_count:
        return
    tries = 0
    while tries < retries:
        print(
            f"retrying validation to check if expected count "
            f"{expected_count} will match current count {current_count}"
        )
        time.sleep(interval)
        current_count = get_index_count(client, index_name)
        if current_count == expected_count:
            break
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from index: {current_count}"
    )


def seed_source_db(df: pd.DataFrame):
    mapping = {
        "properties": {
            "title": {"type": "text", "analyzer": "english"},
            "ethnicity": {"type": "text", "analyzer": "standard"},
            "director": {"type": "text", "analyzer": "standard"},
            "cast": {"type": "text", "analyzer": "standard"},
            "genre": {"type": "text", "analyzer": "standard"},
            "plot": {"type": "text", "analyzer": "english"},
            "year": {"type": "integer"},
            "wiki_page": {"type": "keyword"},
        },
    }
    # seed content
    with get_client() as client:
        client.indices.create(index=SOURCE_INDEX_NAME, mappings=mapping)
        upload_data = dataframe_to_upload_data(df=df)
        bulk(client, upload_data)
        client.indices.refresh(index=SOURCE_INDEX_NAME)
        count = get_index_count(client, SOURCE_INDEX_NAME)
        print(f"seeded {SOURCE_INDEX_NAME} index with {count} records")


@pytest.fixture
def source_index(movies_dataframe: pd.DataFrame) -> str:
    with container_context(
        image="docker.elastic.co/elasticsearch/elasticsearch:8.7.0",
        ports={9200: 9200, 9300: 9300},
        environment={
            "discovery.type": "single-node",
            "xpack.security.enabled": True,
            "ELASTIC_PASSWORD": ES_PASSWORD,
            "ELASTIC_USER": ES_USERNAME,
        },
        healthcheck=HealthCheck(
            test="curl --silent --fail -u ${ELASTIC_USER}:${ELASTIC_PASSWORD} localhost:9200/_cluster/health || exit 1",  # noqa: E501
            interval=1,
            start_period=5,
        ),
    ):
        seed_source_db(df=movies_dataframe)
        yield SOURCE_INDEX_NAME


@pytest.fixture
def destination_index(elasticsearch_elements_mapping: dict) -> str:
    with container_context(
        image="docker.elastic.co/elasticsearch/elasticsearch:8.7.0",
        ports={9200: 9200, 9300: 9300},
        environment={
            "discovery.type": "single-node",
            "xpack.security.enabled": True,
            "ELASTIC_PASSWORD": ES_PASSWORD,
            "ELASTIC_USER": ES_USERNAME,
        },
        healthcheck=HealthCheck(
            test="curl --silent --fail -u ${ELASTIC_USER}:${ELASTIC_PASSWORD} localhost:9200/_cluster/health || exit 1",  # noqa: E501
            interval=1,
            start_period=5,
        ),
    ):
        with get_client() as client:
            response = client.indices.create(
                index=DESTINATION_INDEX_NAME, mappings=elasticsearch_elements_mapping
            )
            if not response["acknowledged"]:
                raise RuntimeError(f"failed to create index: {response}")
        yield DESTINATION_INDEX_NAME


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
async def test_elasticsearch_source(source_index: str, movies_dataframe: pd.DataFrame):
    indexer_config = ElasticsearchIndexerConfig(index_name=source_index)
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        connection_config = ElasticsearchConnectionConfig(
            access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
            username=ES_USERNAME,
            hosts=["http://localhost:9200"],
        )
        download_config = ElasticsearchDownloaderConfig(download_dir=tempdir_path)
        indexer = ElasticsearchIndexer(
            connection_config=connection_config, index_config=indexer_config
        )
        downloader = ElasticsearchDownloader(
            connection_config=connection_config, download_config=download_config
        )
        expected_num_files = len(movies_dataframe)
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=SourceValidationConfigs(
                test_id=CONNECTOR_TYPE,
                expected_num_files=expected_num_files,
                expected_number_indexed_file_data=1,
                validate_downloaded_files=True,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
def test_elasticsearch_source_precheck_fail_no_cluster():
    indexer_config = ElasticsearchIndexerConfig(index_name="index")

    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    indexer = ElasticsearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, NOSQL_TAG)
def test_elasticsearch_source_precheck_fail_no_index(source_index: str):
    indexer_config = ElasticsearchIndexerConfig(index_name="index")

    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    indexer = ElasticsearchIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
async def test_elasticsearch_destination(
    upload_file: Path,
    destination_index: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    stager = ElasticsearchUploadStager(
        upload_stager_config=ElasticsearchUploadStagerConfig(index_name=destination_index)
    )

    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name=destination_index),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with get_client() as client:
        validate_count(client=client, expected_count=expected_count, index_name=destination_index)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
def test_elasticsearch_destination_precheck_fail():
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
def test_elasticsearch_destination_precheck_fail_no_index(destination_index: str):
    connection_config = ElasticsearchConnectionConfig(
        access_config=ElasticsearchAccessConfig(password=ES_PASSWORD),
        username=ES_USERNAME,
        hosts=["http://localhost:9200"],
    )
    uploader = ElasticsearchUploader(
        connection_config=connection_config,
        upload_config=ElasticsearchUploaderConfig(index_name="index"),
    )
    with pytest.raises(DestinationConnectionError):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_elasticsearch_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = ElasticsearchUploadStager(
        upload_stager_config=ElasticsearchUploadStagerConfig(index_name="mock_index")
    )
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import json
from pathlib import Path

import pandas as pd
import pytest

int_test_dir = Path(__file__).parent
assets_dir = int_test_dir / "assets"


@pytest.fixture
def movies_dataframe() -> pd.DataFrame:
    movies_file = assets_dir / "wiki_movie_plots_small.csv"
    assert movies_file.exists()
    assert movies_file.is_file()
    return pd.read_csv(movies_file).dropna().reset_index()


@pytest.fixture
def opensearch_elements_mapping() -> dict:
    elements_mapping_file = assets_dir / "opensearch_elements_mappings.json"
    assert elements_mapping_file.exists()
    assert elements_mapping_file.is_file()
    with elements_mapping_file.open() as fp:
        return json.load(fp)


@pytest.fixture
def elasticsearch_elements_mapping() -> dict:
    elements_mapping_file = assets_dir / "elasticsearch_elements_mappings.json"
    assert elements_mapping_file.exists()
    assert elements_mapping_file.is_file()
    with elements_mapping_file.open() as fp:
        return json.load(fp)



import json
from pathlib import Path

import duckdb
import pytest
from _pytest.fixtures import TopRequest

from test.integration.connectors.utils.constants import DESTINATION_TAG, SQL_TAG
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from unstructured_ingest.v2.processes.connectors.duckdb.duckdb import (
    CONNECTOR_TYPE,
    DuckDBConnectionConfig,
    DuckDBUploader,
    DuckDBUploaderConfig,
    DuckDBUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def provisioned_db_file(duckdb_schema: Path, temp_dir: Path) -> Path:
    db_path = Path(temp_dir) / "temp_duck.db"
    with duckdb.connect(database=db_path) as duckdb_connection:
        with duckdb_schema.open("r") as f:
            query = f.read()
        duckdb_connection.execute(query)
        duckdb_connection.close()
    return db_path


def validate_duckdb_destination(db_path: Path, expected_num_elements: int):
    conn = None
    try:
        conn = duckdb.connect(db_path)
        _results = conn.sql("select count(*) from elements").fetchall()
        _count = _results[0][0]
        assert (
            _count == expected_num_elements
        ), f"dest check failed: got {_count}, expected {expected_num_elements}"
        conn.close()
    finally:
        if conn:
            conn.close()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, "duckdb", SQL_TAG)
def test_duckdb_destination(upload_file: Path, provisioned_db_file: Path, temp_dir: Path):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    stager = DuckDBUploadStager()
    staged_path = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=temp_dir,
        output_filename=upload_file.name,
    )

    connection_config = DuckDBConnectionConfig(database=str(provisioned_db_file))
    upload_config = DuckDBUploaderConfig()
    uploader = DuckDBUploader(connection_config=connection_config, upload_config=upload_config)

    uploader.run(path=staged_path, file_data=file_data)

    with staged_path.open() as f:
        data = json.load(f)
    validate_duckdb_destination(db_path=provisioned_db_file, expected_num_elements=len(data))


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, "duckdb", SQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_duckdb_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = DuckDBUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import asyncio
import json
import os
from pathlib import Path
from typing import Optional

import numpy as np
import pytest
from redis import exceptions as redis_exceptions
from redis.asyncio import Redis, from_url

from test.integration.connectors.utils.constants import DESTINATION_TAG, NOSQL_TAG
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.redisdb import (
    CONNECTOR_TYPE as REDIS_CONNECTOR_TYPE,
)
from unstructured_ingest.v2.processes.connectors.redisdb import (
    RedisAccessConfig,
    RedisConnectionConfig,
    RedisUploader,
    RedisUploaderConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


async def delete_record(client: Redis, element_id: str, key_prefix: str) -> None:
    key_with_prefix = f"{key_prefix}{element_id}"
    await client.delete(key_with_prefix)


async def validate_upload(client: Redis, first_element: dict, key_prefix: str) -> None:
    element_id = first_element["element_id"]
    key_with_prefix = f"{key_prefix}{element_id}"
    expected_text = first_element["text"]
    expected_embeddings = first_element["embeddings"]
    async with client.pipeline(transaction=True) as pipe:
        try:
            response = await pipe.json().get(key_with_prefix, "$").execute()
            response = response[0][0]
        except redis_exceptions.ResponseError:
            response = await pipe.get(key_with_prefix).execute()
            response = json.loads(response[0])

    embedding_similarity = np.linalg.norm(
        np.array(response["embeddings"]) - np.array(expected_embeddings)
    )

    assert response is not None
    assert response["element_id"] == element_id
    assert response["text"] == expected_text
    assert embedding_similarity < 1e-10


async def redis_destination_test(
    upload_file: Path,
    tmp_path: Path,
    connection_kwargs: dict,
    uploader_config: dict,
    uri: Optional[str] = None,
    password: Optional[str] = None,
):
    uploader = RedisUploader(
        connection_config=RedisConnectionConfig(
            **connection_kwargs, access_config=RedisAccessConfig(uri=uri, password=password)
        ),
        upload_config=RedisUploaderConfig(batch_size=10, **uploader_config),
    )
    key_prefix = uploader.upload_config.key_prefix

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=REDIS_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )
    with upload_file.open() as upload_fp:
        elements = json.load(upload_fp)
    first_element = elements[0]

    try:
        if uploader.is_async():
            await uploader.run_data_async(data=elements, file_data=file_data)

        if uri:
            async with from_url(uri) as client:
                await validate_upload(
                    client=client,
                    first_element=first_element,
                    key_prefix=key_prefix,
                )
        else:
            async with Redis(**connection_kwargs, password=password) as client:
                await validate_upload(
                    client=client,
                    first_element=first_element,
                    key_prefix=key_prefix,
                )
    except Exception as e:
        raise e
    finally:
        if uri:
            async with from_url(uri) as client:
                tasks = [
                    delete_record(client, element["element_id"], key_prefix) for element in elements
                ]
                await asyncio.gather(*tasks)
        else:
            async with Redis(**connection_kwargs, password=password) as client:
                tasks = [
                    delete_record(client, element["element_id"], key_prefix) for element in elements
                ]
                await asyncio.gather(*tasks)


@pytest.mark.asyncio
@pytest.mark.tags(REDIS_CONNECTOR_TYPE, DESTINATION_TAG, NOSQL_TAG)
@requires_env("AZURE_REDIS_INGEST_TEST_PASSWORD")
async def test_redis_destination_azure_with_password(upload_file: Path, tmp_path: Path):
    connection_kwargs = {
        "host": "utic-dashboard-dev.redis.cache.windows.net",
        "port": 6380,
        "db": 0,
        "ssl": True,
    }
    uploader_config = {
        "key_prefix": "test_ingest:",
    }
    redis_pw = os.environ["AZURE_REDIS_INGEST_TEST_PASSWORD"]
    await redis_destination_test(
        upload_file, tmp_path, connection_kwargs, uploader_config, password=redis_pw
    )


@pytest.mark.asyncio
@pytest.mark.tags(REDIS_CONNECTOR_TYPE, DESTINATION_TAG, "redis", NOSQL_TAG)
@requires_env("AZURE_REDIS_INGEST_TEST_PASSWORD")
async def test_redis_destination_azure_with_uri(upload_file: Path, tmp_path: Path):
    connection_kwargs = {}
    uploader_config = {
        "key_prefix": "test_ingest:",
    }
    redis_pw = os.environ["AZURE_REDIS_INGEST_TEST_PASSWORD"]
    uri = f"rediss://:{redis_pw}@utic-dashboard-dev.redis.cache.windows.net:6380/0"
    await redis_destination_test(upload_file, tmp_path, connection_kwargs, uploader_config, uri=uri)



import json
import os
import uuid
from contextlib import asynccontextmanager
from pathlib import Path
from typing import AsyncGenerator

import pytest
from _pytest.fixtures import TopRequest
from qdrant_client import AsyncQdrantClient

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.docker import container_context
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.qdrant.cloud import (
    CloudQdrantAccessConfig,
    CloudQdrantConnectionConfig,
    CloudQdrantUploader,
    CloudQdrantUploaderConfig,
    CloudQdrantUploadStager,
    CloudQdrantUploadStagerConfig,
)
from unstructured_ingest.v2.processes.connectors.qdrant.local import (
    CONNECTOR_TYPE as LOCAL_CONNECTOR_TYPE,
)
from unstructured_ingest.v2.processes.connectors.qdrant.local import (
    LocalQdrantConnectionConfig,
    LocalQdrantUploader,
    LocalQdrantUploaderConfig,
    LocalQdrantUploadStager,
    LocalQdrantUploadStagerConfig,
)
from unstructured_ingest.v2.processes.connectors.qdrant.server import (
    CONNECTOR_TYPE as SERVER_CONNECTOR_TYPE,
)
from unstructured_ingest.v2.processes.connectors.qdrant.server import (
    ServerQdrantConnectionConfig,
    ServerQdrantUploader,
    ServerQdrantUploaderConfig,
    ServerQdrantUploadStager,
    ServerQdrantUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

COLLECTION_NAME = f"test-coll-{uuid.uuid4().hex[:12]}"
VECTORS_CONFIG = {"size": 384, "distance": "Cosine"}


@asynccontextmanager
async def qdrant_client(client_params: dict) -> AsyncGenerator[AsyncQdrantClient, None]:
    client = AsyncQdrantClient(**client_params)
    try:
        yield client
    finally:
        await client.close()


async def validate_upload(client: AsyncQdrantClient, upload_file: Path):
    with upload_file.open() as upload_fp:
        elements = json.load(upload_fp)
    expected_point_count = len(elements)
    first_element = elements[0]
    expected_text = first_element["text"]
    embeddings = first_element["embeddings"]
    collection = await client.get_collection(COLLECTION_NAME)
    assert collection.points_count == expected_point_count

    response = await client.query_points(COLLECTION_NAME, query=embeddings, limit=1)
    assert response.points[0].payload is not None
    assert response.points[0].payload["text"] == expected_text


@pytest.mark.asyncio
@pytest.mark.tags(LOCAL_CONNECTOR_TYPE, DESTINATION_TAG, "qdrant", VECTOR_DB_TAG)
async def test_qdrant_destination_local(upload_file: Path, tmp_path: Path):
    connection_kwargs = {"path": str(tmp_path / "qdrant")}
    async with qdrant_client(connection_kwargs) as client:
        await client.create_collection(COLLECTION_NAME, vectors_config=VECTORS_CONFIG)
    AsyncQdrantClient(**connection_kwargs)
    stager = LocalQdrantUploadStager(
        upload_stager_config=LocalQdrantUploadStagerConfig(),
    )
    uploader = LocalQdrantUploader(
        connection_config=LocalQdrantConnectionConfig(**connection_kwargs),
        upload_config=LocalQdrantUploaderConfig(collection_name=COLLECTION_NAME),
    )

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=LOCAL_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    staged_upload_file = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )

    if uploader.is_async():
        await uploader.run_async(path=staged_upload_file, file_data=file_data)
    else:
        uploader.run(path=upload_file, file_data=file_data)
    async with qdrant_client(connection_kwargs) as client:
        await validate_upload(client=client, upload_file=upload_file)


@pytest.fixture
def docker_context():
    with container_context(image="qdrant/qdrant:latest", ports={"6333": "6333"}) as container:
        yield container


@pytest.mark.asyncio
@pytest.mark.tags(SERVER_CONNECTOR_TYPE, DESTINATION_TAG, "qdrant", VECTOR_DB_TAG)
async def test_qdrant_destination_server(upload_file: Path, tmp_path: Path, docker_context):
    connection_kwargs = {"location": "http://localhost:6333"}
    async with qdrant_client(connection_kwargs) as client:
        await client.create_collection(COLLECTION_NAME, vectors_config=VECTORS_CONFIG)
    AsyncQdrantClient(**connection_kwargs)
    stager = ServerQdrantUploadStager(
        upload_stager_config=ServerQdrantUploadStagerConfig(),
    )
    uploader = ServerQdrantUploader(
        connection_config=ServerQdrantConnectionConfig(**connection_kwargs),
        upload_config=ServerQdrantUploaderConfig(collection_name=COLLECTION_NAME),
    )

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=SERVER_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    staged_upload_file = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    if uploader.is_async():
        await uploader.run_async(path=staged_upload_file, file_data=file_data)
    else:
        uploader.run(path=upload_file, file_data=file_data)
    async with qdrant_client(connection_kwargs) as client:
        await validate_upload(client=client, upload_file=upload_file)


@pytest.mark.asyncio
@pytest.mark.tags(SERVER_CONNECTOR_TYPE, DESTINATION_TAG, "qdrant", VECTOR_DB_TAG)
@requires_env("QDRANT_API_KEY", "QDRANT_SERVER_URL")
async def test_qdrant_destination_cloud(upload_file: Path, tmp_path: Path):
    server_url = os.environ["QDRANT_SERVER_URL"]
    api_key = os.environ["QDRANT_API_KEY"]
    connection_kwargs = {"location": server_url, "api_key": api_key}
    async with qdrant_client(connection_kwargs) as client:
        await client.create_collection(COLLECTION_NAME, vectors_config=VECTORS_CONFIG)
    AsyncQdrantClient(**connection_kwargs)

    stager = CloudQdrantUploadStager(
        upload_stager_config=CloudQdrantUploadStagerConfig(),
    )
    uploader = CloudQdrantUploader(
        connection_config=CloudQdrantConnectionConfig(
            url=server_url,
            access_config=CloudQdrantAccessConfig(
                api_key=api_key,
            ),
        ),
        upload_config=CloudQdrantUploaderConfig(collection_name=COLLECTION_NAME),
    )

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=SERVER_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    staged_upload_file = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    if uploader.is_async():
        await uploader.run_async(path=staged_upload_file, file_data=file_data)
    else:
        uploader.run(path=staged_upload_file, file_data=file_data)
    async with qdrant_client(connection_kwargs) as client:
        await validate_upload(client=client, upload_file=upload_file)


@pytest.mark.tags(SERVER_CONNECTOR_TYPE, DESTINATION_TAG, "qdrant", VECTOR_DB_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_qdrant_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = LocalQdrantUploadStager(
        upload_stager_config=LocalQdrantUploadStagerConfig(),
    )
    stager_validation(
        configs=StagerValidationConfigs(test_id=LOCAL_CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import json
import time
import uuid
from datetime import datetime
from pathlib import Path

import pytest
from neo4j import AsyncGraphDatabase, Driver, GraphDatabase
from neo4j.exceptions import ServiceUnavailable
from pytest_check import check

from test.integration.connectors.utils.constants import DESTINATION_TAG, GRAPH_DB_TAG
from test.integration.connectors.utils.docker import container_context
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.utils.chunking import elements_from_base64_gzipped_json
from unstructured_ingest.v2.processes.connectors.neo4j import (
    CONNECTOR_TYPE,
    Label,
    Neo4jAccessConfig,
    Neo4jConnectionConfig,
    Neo4jUploader,
    Neo4jUploaderConfig,
    Neo4jUploadStager,
    Relationship,
)
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

USERNAME = "neo4j"
PASSWORD = "password"
URI = "neo4j://localhost:7687"
DATABASE = "neo4j"

EXPECTED_DOCUMENT_COUNT = 1


# NOTE: Precheck tests are read-only so we utilize the same container for all tests.
# If new tests require clean neo4j container, this fixture's scope should be adjusted.
@pytest.fixture(autouse=True, scope="module")
def _neo4j_server():
    with container_context(
        image="neo4j:latest", environment={"NEO4J_AUTH": "neo4j/password"}, ports={"7687": "7687"}
    ):
        driver = GraphDatabase.driver(uri=URI, auth=(USERNAME, PASSWORD))
        wait_for_connection(driver)
        driver.close()
        yield


@pytest.mark.asyncio
@pytest.mark.tags(DESTINATION_TAG, CONNECTOR_TYPE, GRAPH_DB_TAG)
async def test_neo4j_destination(upload_file: Path, tmp_path: Path):
    stager = Neo4jUploadStager()
    uploader = Neo4jUploader(
        connection_config=Neo4jConnectionConfig(
            access_config=Neo4jAccessConfig(password=PASSWORD),  # type: ignore
            username=USERNAME,
            uri=URI,
            database=DATABASE,
        ),
        upload_config=Neo4jUploaderConfig(),
    )
    file_data = FileData(
        identifier="mock-file-data",
        connector_type="neo4j",
        source_identifiers=SourceIdentifiers(
            filename=upload_file.name,
            fullpath=upload_file.name,
        ),
        metadata=FileDataSourceMetadata(
            date_created=str(datetime(2022, 1, 1).timestamp()),
            date_modified=str(datetime(2022, 1, 2).timestamp()),
        ),
    )
    staged_filepath = stager.run(
        upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )

    await uploader.run_async(staged_filepath, file_data)
    await validate_uploaded_graph(upload_file)

    modified_upload_file = tmp_path / f"modified-{upload_file.name}"
    with open(upload_file) as file:
        elements = json.load(file)
        for element in elements:
            element["element_id"] = str(uuid.uuid4())

    with open(modified_upload_file, "w") as file:
        json.dump(elements, file, indent=4)

    staged_filepath = stager.run(
        modified_upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=modified_upload_file.name,
    )
    await uploader.run_async(staged_filepath, file_data)
    await validate_uploaded_graph(modified_upload_file)


@pytest.mark.tags(DESTINATION_TAG, CONNECTOR_TYPE, GRAPH_DB_TAG)
class TestPrecheck:
    @pytest.fixture
    def configured_uploader(self) -> Neo4jUploader:
        return Neo4jUploader(
            connection_config=Neo4jConnectionConfig(
                access_config=Neo4jAccessConfig(password=PASSWORD),  # type: ignore
                username=USERNAME,
                uri=URI,
                database=DATABASE,
            ),
            upload_config=Neo4jUploaderConfig(),
        )

    def test_succeeds(self, configured_uploader: Neo4jUploader):
        configured_uploader.precheck()

    def test_fails_on_invalid_password(self, configured_uploader: Neo4jUploader):
        configured_uploader.connection_config.access_config.get_secret_value().password = (
            "invalid-password"
        )
        with pytest.raises(
            DestinationConnectionError,
            match="{code: Neo.ClientError.Security.Unauthorized}",
        ):
            configured_uploader.precheck()

    def test_fails_on_invalid_username(self, configured_uploader: Neo4jUploader):
        configured_uploader.connection_config.username = "invalid-username"
        with pytest.raises(
            DestinationConnectionError, match="{code: Neo.ClientError.Security.Unauthorized}"
        ):
            configured_uploader.precheck()

    @pytest.mark.parametrize(
        ("uri", "expected_error_msg"),
        [
            ("neo4j://localhst:7687", "Cannot resolve address"),
            ("neo4j://localhost:7777", "Unable to retrieve routing information"),
        ],
    )
    def test_fails_on_invalid_uri(
        self, configured_uploader: Neo4jUploader, uri: str, expected_error_msg: str
    ):
        configured_uploader.connection_config.uri = uri
        with pytest.raises(DestinationConnectionError, match=expected_error_msg):
            configured_uploader.precheck()

    def test_fails_on_invalid_database(self, configured_uploader: Neo4jUploader):
        configured_uploader.connection_config.database = "invalid-database"
        with pytest.raises(
            DestinationConnectionError, match="{code: Neo.ClientError.Database.DatabaseNotFound}"
        ):
            configured_uploader.precheck()


def wait_for_connection(driver: Driver, retries: int = 10, delay_seconds: int = 2):
    attempts = 0
    while attempts < retries:
        try:
            driver.verify_connectivity()
            return
        except ServiceUnavailable:
            time.sleep(delay_seconds)
            attempts += 1

    pytest.fail("Failed to connect with Neo4j server.")


async def validate_uploaded_graph(upload_file: Path):
    with open(upload_file) as file:
        elements = json.load(file)

    for element in elements:
        if "orig_elements" in element["metadata"]:
            element["metadata"]["orig_elements"] = elements_from_base64_gzipped_json(
                element["metadata"]["orig_elements"]
            )
        else:
            element["metadata"]["orig_elements"] = []

    expected_chunks_count = len(elements)
    expected_element_count = len(
        {
            origin_element["element_id"]
            for chunk in elements
            for origin_element in chunk["metadata"]["orig_elements"]
        }
    )
    expected_nodes_count = expected_chunks_count + expected_element_count + EXPECTED_DOCUMENT_COUNT

    driver = AsyncGraphDatabase.driver(uri=URI, auth=(USERNAME, PASSWORD))
    try:
        nodes_count = len((await driver.execute_query("MATCH (n) RETURN n"))[0])
        chunk_nodes_count = len(
            (await driver.execute_query(f"MATCH (n: {Label.CHUNK.value}) RETURN n"))[0]
        )
        document_nodes_count = len(
            (await driver.execute_query(f"MATCH (n: {Label.DOCUMENT.value}) RETURN n"))[0]
        )
        element_nodes_count = len(
            (await driver.execute_query(f"MATCH (n: {Label.UNSTRUCTURED_ELEMENT.value}) RETURN n"))[
                0
            ]
        )
        with check:
            assert nodes_count == expected_nodes_count
        with check:
            assert document_nodes_count == EXPECTED_DOCUMENT_COUNT
        with check:
            assert chunk_nodes_count == expected_chunks_count
        with check:
            assert element_nodes_count == expected_element_count

        records, _, _ = await driver.execute_query(
            f"""
            MATCH ()-[r:{Relationship.PART_OF_DOCUMENT.value}]->(:{Label.DOCUMENT.value})
            RETURN r
            """
        )
        part_of_document_count = len(records)

        records, _, _ = await driver.execute_query(
            f"""
            MATCH (:{Label.CHUNK.value})-[r:{Relationship.NEXT_CHUNK.value}]->(:{Label.CHUNK.value})
            RETURN r
            """
        )
        next_chunk_count = len(records)

        if not check.any_failures():
            with check:
                assert part_of_document_count == expected_chunks_count + expected_element_count
            with check:
                assert next_chunk_count == expected_chunks_count - 1

    finally:
        await driver.close()



import base64
import hashlib
import json
import zlib
from itertools import groupby


def id_to_hash(element: dict, sequence_number: int) -> str:
    """Calculates and assigns a deterministic hash as an ID.

    The hash ID is based on element's text, sequence number on page,
    page number and its filename.

    Args:
        sequence_number: index on page

    Returns: new ID value
    """
    filename = element["metadata"].get("filename")
    text = element["text"]
    page_number = element["metadata"].get("page_number")
    data = f"{filename}{text}{page_number}{sequence_number}"
    element["element_id"] = hashlib.sha256(data.encode()).hexdigest()[:32]
    return element["element_id"]


def assign_and_map_hash_ids(elements: list[dict]) -> list[dict]:
    # -- generate sequence number for each element on a page --
    elements = elements.copy()
    page_numbers = [e["metadata"].get("page_number") for e in elements]
    page_seq_pairs = [
        seq_on_page for page, group in groupby(page_numbers) for seq_on_page, _ in enumerate(group)
    ]

    # -- assign hash IDs to elements --
    old_to_new_mapping = {
        element["element_id"]: id_to_hash(element=element, sequence_number=seq_on_page_counter)
        for element, seq_on_page_counter in zip(elements, page_seq_pairs)
    }

    # -- map old parent IDs to new ones --
    for e in elements:
        parent_id = e["metadata"].get("parent_id")
        if not parent_id:
            continue
        e["metadata"]["parent_id"] = old_to_new_mapping[parent_id]

    return elements


def elements_from_base64_gzipped_json(raw_s: str) -> list[dict]:
    decoded_b64_bytes = base64.b64decode(raw_s)
    elements_json_bytes = zlib.decompress(decoded_b64_bytes)
    elements_json_str = elements_json_bytes.decode("utf-8")
    element_dicts = json.loads(elements_json_str)
    return element_dicts



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
from time import time
from typing import TYPE_CHECKING, Annotated, Any, Generator, Optional

from dateutil import parser
from pydantic import Field, Secret
from pydantic.functional_validators import BeforeValidator

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import ProviderError, UserAuthError, UserError
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.v2.processes.connectors.utils import conform_string_to_dict
from unstructured_ingest.v2.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileDataSourceMetadata

if TYPE_CHECKING:
    from boxfs import BoxFileSystem

CONNECTOR_TYPE = "box"


class BoxIndexerConfig(FsspecIndexerConfig):
    pass


class BoxAccessConfig(FsspecAccessConfig):
    box_app_config: Annotated[dict, BeforeValidator(conform_string_to_dict)] = Field(
        description="Box app credentials as a JSON string."
    )


class BoxConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["box"], init=False)
    access_config: Secret[BoxAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    def get_access_config(self) -> dict[str, Any]:
        from boxsdk import JWTAuth

        ac = self.access_config.get_secret_value()
        settings_dict = ac.box_app_config

        # Create and authenticate the JWTAuth object
        oauth = JWTAuth.from_settings_dictionary(settings_dict)
        oauth.authenticate_instance()

        # if not oauth.access_token:
        #     raise SourceConnectionError("Authentication failed: No access token generated.")

        # Prepare the access configuration with the authenticated oauth
        access_kwargs_with_oauth: dict[str, Any] = {
            "oauth": oauth,
        }
        access_config: dict[str, Any] = ac.model_dump()
        access_config.pop("box_app_config", None)
        access_kwargs_with_oauth.update(access_config)

        return access_kwargs_with_oauth

    def wrap_error(self, e: Exception) -> Exception:
        from boxsdk.exception import BoxAPIException, BoxOAuthException

        if isinstance(e, BoxOAuthException):
            return UserAuthError(e.message)
        if not isinstance(e, BoxAPIException):
            logger.error(f"unhandled exception from box ({type(e)}): {e}", exc_info=True)
            return e
        message = e.message or e
        if error_code_status := e.status:
            if 400 <= error_code_status < 500:
                return UserError(message)
            if error_code_status >= 500:
                return ProviderError(message)

        logger.error(f"unhandled exception from box ({type(e)}): {e}", exc_info=True)
        return e

    @requires_dependencies(["boxfs"], extras="box")
    @contextmanager
    def get_client(self, protocol: str) -> Generator["BoxFileSystem", None, None]:
        with super().get_client(protocol=protocol) as client:
            yield client


@dataclass
class BoxIndexer(FsspecIndexer):
    connection_config: BoxConnectionConfig
    index_config: BoxIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["name"]
        date_created = None
        date_modified = None
        if modified_at_str := file_info.get("modified_at"):
            date_modified = str(parser.parse(modified_at_str).timestamp())
        if created_at_str := file_info.get("created_at"):
            date_created = str(parser.parse(created_at_str).timestamp())

        file_size = file_info.get("size") if "size" in file_info else None

        version = file_info.get("id")
        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
            "file_id": file_info.get("id"),
        }
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            version=version,
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class BoxDownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class BoxDownloader(FsspecDownloader):
    protocol: str = "box"
    connection_config: BoxConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[BoxDownloaderConfig] = field(default_factory=BoxDownloaderConfig)


class BoxUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class BoxUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: BoxConnectionConfig
    upload_config: BoxUploaderConfig = field(default=None)


box_source_entry = SourceRegistryEntry(
    indexer=BoxIndexer,
    indexer_config=BoxIndexerConfig,
    downloader=BoxDownloader,
    downloader_config=BoxDownloaderConfig,
    connection_config=BoxConnectionConfig,
)

box_destination_entry = DestinationRegistryEntry(
    uploader=BoxUploader,
    uploader_config=BoxUploaderConfig,
    connection_config=BoxConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, TypedDict

from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.processes.uncompress import Uncompressor
from unstructured_ingest.v2.types.file_data import file_data_from_file

STEP_ID = "uncompress"


class UncompressStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class UncompressStep(PipelineStep):
    process: Uncompressor
    identifier: str = STEP_ID

    def __post_init__(self):
        config = self.process.config.model_dump_json() if self.process.config else None
        logger.info(f"created {self.identifier} with configs: {config}")

    async def _run_async(
        self, fn: Callable, path: str, file_data_path: str
    ) -> list[UncompressStepResponse]:
        file_data = file_data_from_file(path=file_data_path)
        fn_kwargs = {"file_data": file_data}
        if not asyncio.iscoroutinefunction(fn):
            new_file_data = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                new_file_data = await fn(**fn_kwargs)
        else:
            new_file_data = await fn(**fn_kwargs)
        responses = []
        for new_file in new_file_data:
            new_file_data_path = Path(file_data_path).parent / f"{new_file.identifier}.json"
            new_file.to_file(path=str(new_file_data_path.resolve()))
            responses.append(
                UncompressStepResponse(
                    path=new_file.local_download_path,
                    file_data_path=str(new_file_data_path),
                )
            )
        return responses



import json
import time
from pathlib import Path

import pytest
import requests
import weaviate
from weaviate.client import WeaviateClient

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.docker import container_context
from unstructured_ingest.v2.processes.connectors.weaviate.local import (
    CONNECTOR_TYPE,
    LocalWeaviateConnectionConfig,
    LocalWeaviateUploader,
    LocalWeaviateUploaderConfig,
    LocalWeaviateUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

COLLECTION_NAME = "elements"


def wait_for_container(timeout: int = 10, interval: int = 1) -> None:
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            requests.get("http://localhost:8080/v1/.well-known/read", timeout=1)
            return
        except Exception as e:
            print(f"Failed to validate container healthy, sleeping for {interval} seconds: {e}")
            time.sleep(interval)
    raise TimeoutError("Docker container never came up healthy")


@pytest.fixture
def weaviate_instance():
    with container_context(
        image="semitechnologies/weaviate:1.27.3",
        ports={8080: 8080, 50051: 50051},
    ) as ctx:
        wait_for_container()
        yield ctx


@pytest.fixture
def collection(weaviate_instance, collections_schema_config: dict) -> str:
    with weaviate.connect_to_local() as weaviate_client:
        weaviate_client.collections.create_from_dict(config=collections_schema_config)
    return COLLECTION_NAME


def get_count(client: WeaviateClient) -> int:
    collection = client.collections.get(COLLECTION_NAME)
    resp = collection.aggregate.over_all(total_count=True)
    return resp.total_count


def validate_count(expected_count: int, retries: int = 10, interval: int = 1) -> None:
    with weaviate.connect_to_local() as weaviate_client:
        current_count = get_count(client=weaviate_client)
        retry_count = 0
        while current_count != expected_count and retry_count < retries:
            retry_count += 1
            time.sleep(interval)
            current_count = get_count(client=weaviate_client)
        assert current_count == expected_count, (
            f"Expected count ({expected_count}) doesn't match how "
            f"much came back from collection: {current_count}"
        )


def run_uploader_and_validate(
    uploader: LocalWeaviateUploader, path: Path, file_data: FileData, expected_count: int
):
    uploader.precheck()
    uploader.run(path=path, file_data=file_data)
    validate_count(expected_count=expected_count)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_weaviate_local_destination(upload_file: Path, collection: str, tmp_path: Path):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    stager = LocalWeaviateUploadStager()

    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    dynamic_uploader = LocalWeaviateUploader(
        upload_config=LocalWeaviateUploaderConfig(
            collection=COLLECTION_NAME,
        ),
        connection_config=LocalWeaviateConnectionConfig(),
    )
    fixed_size_uploader = LocalWeaviateUploader(
        upload_config=LocalWeaviateUploaderConfig(
            collection=COLLECTION_NAME, batch_size=10, dynamic_batch=False
        ),
        connection_config=LocalWeaviateConnectionConfig(),
    )
    rate_limited_uploader = LocalWeaviateUploader(
        upload_config=LocalWeaviateUploaderConfig(
            collection=COLLECTION_NAME, requests_per_minute=50, dynamic_batch=False
        ),
        connection_config=LocalWeaviateConnectionConfig(),
    )
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)

    run_uploader_and_validate(
        uploader=dynamic_uploader,
        path=staged_filepath,
        file_data=file_data,
        expected_count=expected_count,
    )
    run_uploader_and_validate(
        uploader=fixed_size_uploader,
        path=staged_filepath,
        file_data=file_data,
        expected_count=expected_count,
    )
    run_uploader_and_validate(
        uploader=rate_limited_uploader,
        path=staged_filepath,
        file_data=file_data,
        expected_count=expected_count,
    )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_weaviate_local_create_destination(weaviate_instance):
    uploader = LocalWeaviateUploader(
        upload_config=LocalWeaviateUploaderConfig(),
        connection_config=LocalWeaviateConnectionConfig(),
    )
    collection_name = "system_created-123"
    formatted_collection_name = "System_created_123"
    created = uploader.create_destination(destination_name=collection_name)
    assert created
    with uploader.connection_config.get_client() as weaviate_client:
        assert weaviate_client.collections.exists(name=formatted_collection_name)

    created = uploader.create_destination(destination_name=collection_name)
    assert not created
